#define MAXSNDLEN 32768
#define ADPCMMIN -4
#define ADPCMMAX 3

static unsigned char snd[MAXSNDLEN], sndcomp[MAXSNDLEN];
static long sndbytelen, sndbitlen;
static unsigned char astart0, astart1, astart2;
static unsigned char hufflen[8] = {5,5,4,3,1,3,4,4};
static unsigned char huffval[8] = {0xf,0x1f,3,1,0,5,0xb,7};

#pragma aux resetsb =\
	"mov al, 1",\
	"mov dx, 0x220",\
	"add dx, 0x6",\
	"out dx, al",\
	"in al, dx",\
	"in al, dx",\
	"in al, dx",\
	"in al, dx",\
	"xor al, al",\
	"out dx, al",\
	"add dx, 4",\
	"mov cx, 100",\
	"resetsb1: in al, dx",\
	"cmp al, 0xaa",\
	"je resetsb2",\
	"loop resetsb1",\
	"mov ax, 1",\
	"jmp resetsb3",\
	"resetsb2: xor ax, ax",\
	"resetsb3:",\
	modify [eax ecx edx]\

#pragma aux portwait =\
	"beginwait: in al, dx",\
	"xor al, cl",\
	"test al, bl",\
	"jz beginwait",\
	parm [edx][ebx][ecx]\

#pragma aux huffputbits =\
	"xor edx, edx",\
	"mov dl, byte ptr huffval[eax]",\
	"mov ecx, sndbitlen",\
	"and ecx, 7",\
	"shl edx, cl",\
	"mov ecx, sndbitlen",\
	"shr ecx, 3",\
	"or dword ptr sndcomp[ecx], edx",\
	"xor ebx, ebx",\
	"mov bl, byte ptr hufflen[eax]",\
	"add dword ptr sndbitlen, ebx",\
	parm [eax]\
	modify [ebx ecx edx]\

main()
{
	long i;
	char ch;

	resetsb();

	outp(0x43,0x92);  //Initializes timer (IMPORTANT for this POLLING method)

	printf("Hold down a shift key to start recording\n");
	record();

	printf("       Raw sound: %ld bits\n",sndbytelen<<3);
	printf("Compressed sound: %ld bits\n",sndbitlen);

	printf("Playback...\n");
	playback();
}

lopassfilter()
{
	long z;

	for(z=0;z<sndbytelen;z++)
		sndcomp[z] = snd[z];
	for(z=2;z<sndbytelen-2;z++)
		snd[z] = (sndcomp[z-1] + sndcomp[z]*3 + sndcomp[z+1])/5;
}
     
record()
{
	long i;

	outp(0x21,inp(0x21)|3);
	portwait(0x42,64L,0L);
	portwait(0x42,64L,255L);
	printf("Recording...\n");
	i = 0;
	while ((i < MAXSNDLEN) && (kbhit() == 0))
	{
		portwait(0x42,64L,0L);
		outp(0x22c,0x20);
		portwait(0x42,64L,255L);
		snd[i++] = inp(0x22a);
	}
	outp(0x21,inp(0x21)&0xfc);
	lopassfilter();
	adpcmencode();
}

adpcmencode()
{
	long z, pred, dat, scale, a0, a1, a2;

	for(z=0;z<sndbytelen;z++)
		sndcomp[z] = 0;
	sndbitlen = 0;
	astart0 = snd[0]; astart1 = snd[1]; astart2 = snd[1];
	a0 = astart0; a1 = astart1; a2 = astart2;
	scale = 2;
	for(z=3;z<sndbytelen;z++)
	{
		pred = a2 + (((a2-a1) + ((a2-a1) - (a1-a0)))>>1);
		if (pred < 0) pred = 0;
		if (pred > 255) pred = 255;
		dat = (snd[z] - pred) / scale;
		if (dat > ADPCMMAX) dat = ADPCMMAX;
		if (dat < ADPCMMIN) dat = ADPCMMIN;
		a0 = a1; a1 = a2; a2 = pred + dat * scale;
		huffputbits((long)(dat-ADPCMMIN));
		if ((dat == ADPCMMIN) || (dat == ADPCMMAX)) scale <<= 1;
		if ((dat == 0) && (scale > 2)) scale >>= 1;
	}
}

adpcmdecode()
{
	long z, pred, dat, scale, a0, a1, a2;

	sndbitlen = 0;
	snd[0] = astart0; snd[1] = astart1; snd[2] = astart2; scale = 2;
	for(z=3;z<sndbytelen;z++)
	{
		dat = huffgetbits();
		dat += ADPCMMIN;
		a0 = snd[z-3]; a1 = snd[z-2]; a2 = snd[z-1];
		pred = a2 + (((a2-a1) + ((a2-a1) - (a1-a0)))>>1);
		if (pred < 0) pred = 0;
		if (pred > 255) pred = 255;
		snd[z] = pred + dat * scale;
		if ((dat == ADPCMMIN) || (dat == ADPCMMAX)) scale <<= 1;
		if ((dat == 0) && (scale > 2)) scale >>= 1;
	}
}

huffgetbits()
{
	long i, j, dat;

	dat = (long)sndcomp[sndbitlen>>3]+((long)sndcomp[(sndbitlen>>3)+1]<<8);
	dat >>= (sndbitlen&7);
	for(i=0;i<8;i++)
	{
		if ((dat&((1<<hufflen[i])-1)) == ((long)huffval[i]))
		{
			j = i;
			break;
		}
	}

	sndbitlen += hufflen[j];
	return(j);
}

playback()
{
	long i;

	adpcmdecode();
	lopassfilter();
	outp(0x21,inp(0x21)|3);
	portwait(0x22c,128L,255); outp(0x22c,0xd1);
	portwait(0x42,64L,0L);
	portwait(0x42,64L,255L);
	for(i=0;i<sndbytelen;i++)
	{
		portwait(0x42,64L,0L);
		outp(0x22c,0x10);
		portwait(0x42,64L,255L);
		outp(0x22c,snd[i]);
	}
	portwait(0x22c,128L,255); outp(0x22c,0xd3);
	outp(0x21,inp(0x21)&0xfc);
}
