// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXSECTORS 1024
#define MAXWALLS 4096
#define MAXSPRITES 4096

static struct sectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingosc, ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorosc, floorshade;
	long floorz;
	short floorpicnum, floorheinum;
	long tag;
};

static struct walltype
{
	long x, y;
	short point2;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	long tag;
};

static struct spritetype
{
	long x, y, z;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
	short sectnum, statnum;
	long tag;
};

//**************************************************************************

static struct sectortype sector[MAXSECTORS];
static struct walltype wall[MAXWALLS];
static struct spritetype sprite[MAXSPRITES];

static long posx, posy, posz, mapversion;
static short int ang, cursectnum, numsectors, numwalls, numsprites;

main(int argc, char **argv)
{
	char filename[80];
	long i, j;

	if (argc != 2)
	{
		printf("CONVMAP [filename]\n");
		printf("This CONVMAP.EXE converts map version 0 to map version 1\n");
		exit(0);
	}
	strcpy(filename,argv[1]);
	if(strchr(filename,'.') == 0)
		strcat(filename,".map");

	loadoldboard(filename);
	savenewboard(filename);

	if ((i = loadoldboard(filename)) < 0)
	{
		if (i == -1) printf("%s not found.\n",filename);
		if (i == -2) printf("Cannot convert %s.\n",filename);
		exit(0);
	}

	printf("Converting bit 2 of sector ceilingstat (doors) to tags...\n");
	printf("What sector tag # do you want the doors to converted to?\n");
	scanf("%ld",&j);
	printf("Doors converted to tag #%ld\n",j);

	for(i=0;i<numsectors;i++)
	{
		if ((sector[i].ceilingstat&4) > 0)
		{
			if (sector[i].tag > 0)
				printf("Warning: Overwriting tag %d\n",sector[i].tag&65535);
			sector[i].tag = (sector[i].tag&0xffff0000) + (j&65535);
			sector[i].ceilingstat &= (255-4);
		}
	}
	i = 0;
	while (i < numsprites)
	{
		if ((sprite[i].x == 0) && (sprite[i].y == 0) && (sprite[i].z == 0))
		{
			if ((sprite[i].xrepeat == 0) && (sprite[i].yrepeat == 0))
			{
				if ((sprite[i].picnum == 0) && (sprite[i].sectnum == 0))
				{
					numsprites--;
					if (i != numsprites)
						memcpy(&sprite[i],&sprite[numsprites],sizeof(sprite[0]));
				}
				else
				{
					i++;
				}
			}
			else
			{
				i++;
			}
		}
		else
		{
			i++;
		}
	}


	mapversion = 1L;
	if (savenewboard(filename) == 0)
		printf("%s successfully converted.\n",filename);
	else
		printf("%s NOT converted for some stupid reason.\n",filename);
}

loadoldboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&mapversion,4);
	if (mapversion != 0L) return(-2);
	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&cursectnum,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) read(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) read(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) read(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);
	return(0);
}

savenewboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);

	write(fil,&mapversion,4);
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&cursectnum,2);
	write(fil,&numsectors,2);
	write(fil,&numwalls,2);
	write(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) write(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) write(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) write(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);
	return(0);
}
