// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXSECTORS 1024
#define MAXWALLS 4096
#define MAXSPRITES 4096

//**************************************************************************

static struct sectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingosc, ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorosc, floorshade;
	long floorz;
	short floorpicnum, floorheinum;
};

static struct walltype
{
	long x, y;
	short point2;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
};

static struct spritetype
{
	long x, y, z;
	short x2offs, y2offs;
	short point2;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
};

//**************************************************************************

static struct newsectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingosc, ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorosc, floorshade;
	long floorz;
	short floorpicnum, floorheinum;
	long tag;
};

static struct newwalltype
{
	long x, y;
	short point2;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	long tag;
};

static struct newspritetype
{
	long x, y, z;
	char cstat, shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
	short sectnum, statnum;
	long tag;
};

//**************************************************************************

static struct sectortype sector[MAXSECTORS];
static struct walltype wall[MAXWALLS];
static struct spritetype sprite[MAXSPRITES];

static struct newsectortype newsector[MAXSECTORS];
static struct newwalltype newwall[MAXWALLS];
static struct newspritetype newsprite[MAXSPRITES];

static long posx, posy, posz, versionum;
static short int ang, cursectnum, numsprites, numsectors, numwalls;
static short firstsprite[MAXSECTORS], spritesectnum[MAXSPRITES];

main(int argc, char **argv)
{
	char filename[80];
	long i;

	if (argc != 2)
	{
		printf("CONVMAP [filename]\n");
		exit(0);
	}
	strcpy(filename,argv[1]);
	if(strchr(filename,'.') == 0)
		strcat(filename,".map");

	if ((i = loadoldboard(filename)) < 0)
	{
		if (i == -1) printf("%s not found.\n",filename);
		if (i == -2) printf("%s already updated.\n",filename);
		exit(0);
	}

	versionum = 0L;
	for(i=0;i<numsectors;i++)
	{
		newsector[i].wallptr = sector[i].wallptr;
		newsector[i].wallnum = sector[i].wallnum;
		newsector[i].ceilingstat = sector[i].ceilingstat;
		newsector[i].ceilingosc = sector[i].ceilingosc;
		newsector[i].ceilingshade = sector[i].ceilingshade;
		newsector[i].ceilingz = sector[i].ceilingz;
		newsector[i].ceilingpicnum = sector[i].ceilingpicnum;
		newsector[i].ceilingheinum = sector[i].ceilingheinum;
		newsector[i].floorstat = sector[i].floorstat;
		newsector[i].floorosc = sector[i].floorosc;
		newsector[i].floorshade = sector[i].floorshade;
		newsector[i].floorz = sector[i].floorz;
		newsector[i].floorpicnum = sector[i].floorpicnum;
		newsector[i].floorheinum = sector[i].floorheinum;
		newsector[i].tag = 0L;
	}
	for(i=0;i<numwalls;i++)
	{
		newwall[i].x = wall[i].x;
		newwall[i].y = wall[i].y;
		newwall[i].point2 = wall[i].point2;
		newwall[i].cstat = wall[i].cstat;
		newwall[i].shade = wall[i].shade;
		newwall[i].xrepeat = wall[i].xrepeat;
		newwall[i].yrepeat = wall[i].yrepeat;
		newwall[i].picnum = wall[i].picnum;
		newwall[i].nextsector1 = wall[i].nextsector1;
		newwall[i].nextwall1 = wall[i].nextwall1;
		newwall[i].nextsector2 = wall[i].nextsector2;
		newwall[i].nextwall2 = wall[i].nextwall2;
		newwall[i].tag = 0L;
	}
	for(i=0;i<numsprites;i++)
	{
		newsprite[i].x = sprite[i].x;
		newsprite[i].y = sprite[i].y;
		newsprite[i].z = sprite[i].z;
		newsprite[i].cstat = sprite[i].cstat;
		newsprite[i].shade = sprite[i].shade;
		newsprite[i].xrepeat = sprite[i].xrepeat;
		newsprite[i].yrepeat = sprite[i].yrepeat;
		newsprite[i].picnum = sprite[i].picnum;
		newsprite[i].ang = sprite[i].ang;
		newsprite[i].vel = sprite[i].vel;
		newsprite[i].zvel = sprite[i].zvel;
		newsprite[i].owner = sprite[i].owner;
		newsprite[i].sectnum = spritesectnum[i];
		newsprite[i].statnum = 0;
		newsprite[i].tag = 0L;
	}

	savenewboard(filename);
	printf("%s successfully converted.\n",filename);
}

loadoldboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&posx,4);
	if (posx == 0L)
		return(-2);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&numsprites,2);
	read(fil,&cursectnum,2);
	read(fil,&firstsprite[0],(numsectors<<1));
	read(fil,&spritesectnum[0],(numsprites<<1));
	for(i=0;i<numsectors;i++) read(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) read(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) read(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);

	return(0);
}

savenewboard(char filename[20])
{
	short int fil, i, j;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);
	write(fil,&versionum,4);
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&cursectnum,2);
	write(fil,&numsectors,2);
	write(fil,&numwalls,2);
	write(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) write(fil,&newsector[i],sizeof(struct newsectortype));
	for(i=0;i<numwalls;i++) write(fil,&newwall[i],sizeof(struct newwalltype));
	for(i=0;i<numsprites;i++) write(fil,&newsprite[i],sizeof(struct newspritetype));
	close(fil);
	return(0);
}
