#include <dos.h>
#include <malloc.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include "pragmas.h"

#define MAXXDIM 1600L
#define MAXYDIM 1200L
#define BSIZ 16
#define WALLDIST 128
#define MAXPOINTS 4096
#define MAXPOLYS 1024
#define MAXSPLITCNT 1024
#define MAXPALOOKUPS 64
#define MAXTILES 4096
#define MAXTILEFILES 256
#define NUMOPTIONS 8
#define NUMKEYS 19

#define AVERAGEFRAMES 16
static long frameval[AVERAGEFRAMES], framecnt = 0;

#define reduce(x,y,z,val)                                  \
{                                                          \
	while (klabs(x) >= (val<<3)) x >>= 4, y >>= 4, z >>= 4; \
	while (klabs(y) >= (val<<3)) x >>= 4, y >>= 4, z >>= 4; \
	while (klabs(z) >= (val<<3)) x >>= 4, y >>= 4, z >>= 4; \
	while (klabs(x) >= val) x >>= 1, y >>= 1, z >>= 1;      \
	while (klabs(y) >= val) x >>= 1, y >>= 1, z >>= 1;      \
	while (klabs(z) >= val) x >>= 1, y >>= 1, z >>= 1;      \
}

static long gxx, gxy, gxz, gyx, gyy, gyz, gzx, gzy, gzz;       //rotation

#define rotate3d(nx,ny,nz)                                          \
{                                                                   \
	long rotate3dox = nx, rotate3doy = ny, rotate3doz = nz;          \
	nx = tmulscale28(rotate3dox,gxx,rotate3doy,gyx,rotate3doz,gzx);  \
	ny = tmulscale28(rotate3dox,gxy,rotate3doy,gyy,rotate3doz,gzy);  \
	nz = tmulscale28(rotate3dox,gxz,rotate3doy,gyz,rotate3doz,gzz);  \
}

#define backrotate3d(nx,ny,nz)                                      \
{                                                                   \
	long rotate3dox = nx, rotate3doy = ny, rotate3doz = nz;          \
	nx = tmulscale28(rotate3dox,gxx,rotate3doy,gxy,rotate3doz,gxz);  \
	ny = tmulscale28(rotate3dox,gyx,rotate3doy,gyy,rotate3doz,gyz);  \
	nz = tmulscale28(rotate3dox,gzx,rotate3doy,gzy,rotate3doz,gzz);  \
}

typedef struct
{
	long x, y;
} point2d;

typedef struct
{
	long x, y, z;
} point3d;

	//Board layout
static short board[6*BSIZ*BSIZ*BSIZ];

	//6 cube planes in clockwise order
static long of[6*4*3] =
{
	1,1,1,1,1,-1,1,-1,-1,1,-1,1,     //posx > x
	-1,1,-1,1,1,-1,1,1,1,-1,1,1,     //posy > y
	-1,1,1,1,1,1,1,-1,1,-1,-1,1,     //posz > z
	-1,1,-1,-1,1,1,-1,-1,1,-1,-1,-1, //posx < x
	-1,-1,1,1,-1,1,1,-1,-1,-1,-1,-1, //posy < y
	1,1,-1,-1,1,-1,-1,-1,-1,1,-1,-1, //posz < z
};

	//positions/angles and their velocities
static long posx, posy, posz, a1, a2, a3;
static long fvel = 0, svel = 0, hvel = 0, a1vel = 0, a2vel = 0, a3vel = 0;
  
	//New polygon variables
static point3d r[MAXPOINTS];
static point2d sc[MAXPOINTS];
static short point2[MAXPOINTS], numpoints, poly[MAXPOLYS], numpolys;
static short plist[MAXPOINTS];

	//beam tree variables
static point3d bn[MAXPOINTS];
static short front[MAXPOINTS], back[MAXPOINTS], pnum;
static short pt[MAXPOINTS], polytag[MAXPOLYS], splittag[MAXPOINTS];
static short split[MAXSPLITCNT], bak[MAXPOLYS];

static short cubelist[BSIZ*BSIZ*BSIZ];

static long arealeft;
static long globalregioncnt, globalsplitcnt;

	//Diagonal filling variables
long hplc[MAXXDIM+1], hinc[MAXXDIM+1], ui[4], vi[4], pinc[4];
char hind[MAXXDIM+1];
static short h1[(MAXXDIM<<1)+4], n1[MAXXDIM<<5], p1[MAXXDIM<<5];
static short h2[(MAXXDIM<<1)+4], n2[MAXXDIM<<5], p2[MAXXDIM<<5];

static long gm, gd, gdx, gdy, gu, gux, guy, gv, gvx, gvy, gt;  //texturizing
static long gxoffs, gyoffs;

static char *gbuf;
static long gshade, gtext, gmask1, gmask2;

static char *screen;
static long frameplace, ylookup[MAXYDIM], zoom, xreszoom, ydimscale;
static long xdim, ydim, visualpage = 0;

static short mousx, mousy, bstatus, obstatus, moustat;
static long searchx, searchy;
static char intbstatus, mouseirq, mousetemp;
static char mousebyte[3], mouselookup[4] = {0,2,1,3}, mousebytecnt;
static long intmousx, intmousy, mouseport = -1;
static void (__interrupt __far *oldmousehandler)();
static void __interrupt __far mousehandler(void);

static long vesares[7][2] = {320,200,640,400,640,480,800,600,1024,768,
									  1280,1024,1600,1200};

static char vesapageshift, curpag, tsengdriver;

	//Palette variables
static char paletteloaded = 0, translucloaded = 0;
static short numpalookups;
static char palette[768], whitecol, blackcol;
static char palookup[MAXPALOOKUPS][256], *transluc;

	//Tiles???.ART variables
static char *pic, picsiz[MAXTILES];        //pic-pointer to BIG tile buffer
static long numtiles, waloff[MAXTILES], picanm[MAXTILES];  //waloff-offset
static long tilesizx[MAXTILES], tilesizy[MAXTILES];    //dimensions of tile
	//Internal art variables
static char artfilename[20];
static long artsize = 0, artversion, picsloaded = 0;
static long artfil, artfilnum, artfilplc, numtilefiles;
static char tilefilenum[MAXTILES];
static long tilefileoffs[MAXTILES];
static long localtilestart[MAXTILEFILES], localtileend[MAXTILEFILES];

	//Temp variables - Save memory by using these
static char tempbuf[4096];
static short tempshort[MAXTILES];

	//Editor variables (tab&enter)
char boardfilename[13];
long temppicnum = -1, tempshade = -1, dragpoint = -1;
short searchcube, searchcubak, searchedge;
long searchitx, searchity, searchitz;

static char palette[768];

static short sintable[2048];
static long sqrtable[2048];
static char britable[16][64];
static char textfont[1024], smalltextfont[1024];
char option[NUMOPTIONS], keys[NUMKEYS];
long randomseed;

long totalclock, baktotalclock, lockspeed, numframes;
static void (__interrupt __far *oldtimerhandler)();
static void __interrupt __far timerhandler(void);

volatile char keystatus[256], readch, oldreadch, extended, keytemp;
static void (__interrupt __far *oldkeyhandler)();
static void __interrupt __far keyhandler(void);

extern long setupdline(long,long,long);
#pragma aux setupdline parm [eax][ebx][ecx];
extern long dline(long,long,long,long,long,long);
#pragma aux dline parm [eax][ebx][ecx][edx][esi][edi];

// 11,12 is 2,1-10
// 13,14 is 4,2-10
// 15,16 is 6,3-10
#pragma aux ksqrtasm =\
	"or eax, eax",\
	"jz endksqrtasm",\
	"jns skipnegate",\
	"neg eax",\
	"skipnegate: bsr ecx, eax",\
	"cmp ecx, 11",\
	"jge over2048",\
	"mov eax, dword ptr sqrtable[eax*4]",\
	"shr eax, 10",\
	"jmp endksqrtasm",\
	"over2048:",\
	"sub cl, 9",\
	"and cl, 0xfe",\
	"shr eax, cl",\
	"mov eax, dword ptr sqrtable[eax*4]",\
	"shr cl, 1",\
	"shl eax, cl",\
	"shr eax, 10",\
	"endksqrtasm:",\
	parm [eax]\
	modify [eax ebx ecx edx]\

#pragma aux vesasetmode =\
	"mov ax, 0x4f02",\
	"int 0x10",\
	"and eax, 0x0000ffff",\
	parm [ebx]\
	modify [eax ebx]\

#pragma aux vesasetpage =\
	"mov cl, vesapageshift",\
	"shl dx, cl",\
	"cmp dl, curpag",\
	"je skipsetpage",\
	"mov curpag, dl",\
	"mov ax, 0x4f05",\
	"xor bx, bx",\
	"int 0x10",\
	"skipsetpage:",\
	parm [edx]\
	modify [eax ebx ecx]\

#pragma aux setupandgetmouseirq =\
	"xor eax, eax",\
	"int 0x33",\
	"jz skipit",\
	"mov eax, 36",\
	"int 0x33",\
	"mov mouseirq, cl",\
	"mov eax, -1",\
	"skipit:",\
	modify [eax ecx]\

void __interrupt __far timerhandler()
{
	totalclock++;
	keytimerstuff();

	_chain_intr(oldtimerhandler);
	koutp(0x20,0x20);
}

keytimerstuff()
{
		//Acceleration
	if (keystatus[keys[0]]) fvel = min(fvel+8,127);
	if (keystatus[keys[1]]) fvel = max(fvel-8,-128);
	if (keystatus[0x2c]) hvel = min(hvel+8,127);
	if (keystatus[0x1e]) hvel = max(hvel-8,-128);
	if (keystatus[keys[5]])
	{
		if (keystatus[keys[3]]) svel = min(svel+8,127);
		if (keystatus[keys[2]]) svel = max(svel-8,-128);
	}
	else
	{
		if (keystatus[keys[3]]) a3vel = min(a3vel+16,127);
		if (keystatus[keys[2]]) a3vel = max(a3vel-16,-128);
	}
	if (keystatus[0x33]) a1vel = min(a1vel+32,127);
	if (keystatus[0x34]) a1vel = max(a1vel-32,-128);
	if (keystatus[keys[11]]) a2vel = min(a2vel+16,127);
	if (keystatus[keys[10]]) a2vel = max(a2vel-16,-128);

		//Friction
	if (fvel < 0) fvel = min(fvel+2,0);
	if (fvel > 0) fvel = max(fvel-2,0);
	if (svel < 0) svel = min(svel+2,0);
	if (svel > 0) svel = max(svel-2,0);
	if (hvel < 0) hvel = min(hvel+2,0);
	if (hvel > 0) hvel = max(hvel-2,0);
	if (a1vel < 0) a1vel = min(a1vel+12,0);
	if (a1vel > 0) a1vel = max(a1vel-12,0);
	if (a2vel < 0) a2vel = min(a2vel+12,0);
	if (a2vel > 0) a2vel = max(a2vel-12,0);
	if (a3vel < 0) a3vel = min(a3vel+12,0);
	if (a3vel > 0) a3vel = max(a3vel-12,0);
}

void __interrupt __far keyhandler()
{
	oldreadch = readch; readch = kinp(0x60);
	keytemp = kinp(0x61); koutp(0x61,keytemp|128); koutp(0x61,keytemp&127);
	if ((readch|1) == 0xe1)
		extended = 128;
	else
	{
		if (oldreadch != readch)
			keystatus[(readch&127)+extended] = ((readch>>7)^1);
		extended = 0;
	}
	koutp(0x20,0x20);
	if (keystatus[0x46])
	{
		koutp(0x43,54); koutp(0x40,255); koutp(0x40,255);
		_dos_setvect(0x9, oldkeyhandler);
		_dos_setvect(0x8, oldtimerhandler);
		free(screen);
		setvmode(0x3);
		exit(0);
	}
}

main()
{
	long i, j, x, y, z, xvect, yvect, zvect;
	char shiftstate;

	loadtables();

	xdim = 320; ydim = 200;
	if (option[0] == 1)
	{
		xdim = vesares[option[6]&15][0];
		ydim = vesares[option[6]&15][1];
	}
	searchx = (xdim>>1);
	searchy = (ydim>>1);

	if (option[3] == 1) initmouse();
	loadpics("tiles000.art");
	loadpalette();

	switch(option[0])
	{
		case 0: case 1:
			if ((screen = (char *)malloc(xdim*ydim)) == NULL)
			{
				setvmode(0x3);
				printf("Not enough memory for screen buffer\n");
				exit(0);
			}
			frameplace = FP_OFF(screen);

			i = -1;
			if ((xdim == 320) && (ydim == 200)) setvmode(0x13); i = 1;
			if ((xdim == 640) && (ydim == 400)) i = setvesamode(0x100);
			if ((xdim == 640) && (ydim == 480)) i = setvesamode(0x101);
			if ((xdim == 800) && (ydim == 600)) i = setvesamode(0x103);
			if ((xdim == 1024) && (ydim == 768)) i = setvesamode(0x105);
			if ((xdim == 1280) && (ydim == 1024)) i = setvesamode(0x107);
			if ((xdim == 1600) && (ydim == 1200)) i = setvesamode(0x120);
			if (i < 0)
			{
				setvmode(0x3);
				printf("%ld * %ld Vesa mode not supported\n",xdim,ydim);
				exit(0);
			}
			break;
		case 2:
			setvmode(0x13);
			frameplace = 0xa0000;
			visualpage = 0; koutp(0x3cd,visualpage+(visualpage<<4));
			break;
		case 3:
			setvmode(0x13);
			koutp(0x3ce,0xf); koutp(0x3cf,0x5);
			koutp(0x3d4,0x29); koutp(0x3d5,0x85);
			koutp(0x3c4,0x6); koutp(0x3c5,0x48);
			koutp(0x3d4,0x2f); koutp(0x3d5,0);   //set mod = 256k
			frameplace = 0xa0000; visualpage = 0; koutp(0x3ce,0x9); koutp(0x3cf,visualpage<<4);
			break;
		case 4:
			setvmode(0x13);
			koutp(0x3d4,0x38); koutp(0x3d5,0x48);           //S3 Extensions enable #1
			koutp(0x3d4,0x39); koutp(0x3d5,0xa5);           //S3 Extensions enable #2
			koutp(0x3d4,0x31); koutp(0x3d5,kinp(0x3d5)|9);  //Init for start address

			frameplace = 0xa0000;
			visualpage = 0; koutp(0x3d4,0x35); koutp(0x3d5,visualpage);
			break;
	}

	koutp(0x3c8,0);
	for(i=0;i<768;i++) koutp(0x3c9,palette[i]);

	j = 0;
	for(i=0;i<MAXYDIM;i++) ylookup[i] = j, j += xdim;

	srand(readtimer());

	loadboard("board.cub");

	oldtimerhandler = _dos_getvect(0x8);
	_dos_setvect(0x8, timerhandler);
	koutp(0x43,54); koutp(0x40,9942&255); koutp(0x40,9942>>8);
	oldkeyhandler = _dos_getvect(0x9);
	_disable(); _dos_setvect(0x9, keyhandler); _enable();
	totalclock = 0L;
	baktotalclock = 0L;
	numframes = 0;
	lockspeed = 0;

	zoom = 65536;
	xreszoom = mulscale(zoom,xdim,1);
	ydimscale = divscale32(ydim,xdim*zoom);

	while (keystatus[1] == 0)
	{
		setuprotate3d();

		drawscreen();

		shiftstate = (keystatus[0x2a]|keystatus[0x36]);

		xvect = (svel<<(shiftstate+1));
		yvect = (hvel<<(shiftstate+0));
		zvect = (fvel<<(shiftstate+1));
		if ((xvect|yvect|zvect) != 0)
		{
			xvect *= (lockspeed<<7);
			yvect *= (lockspeed<<7);
			zvect *= (lockspeed<<7);
			backrotate3d(xvect,yvect,zvect);
			clipmove(&posx,&posy,&posz,xvect>>11,yvect>>11,zvect>>11);
		}

		if (keystatus[0x52] > 0)    //0 on keypad (return to normal angle)
		{
			if ((keystatus[0x1d]|keystatus[0x9d]) > 0)
			{
				a2 = ((a2+1536)&2047);
				if (a2 < 1024) { a2 -= (lockspeed<<1); if (a2 < 0) a2 = 0; }
				else           { a2 += (lockspeed<<1); if (a2 >= 2048) a2 = 0; }
				a2 = ((a2+512)&2047);
			}
			else
			{
				if (a2 < 1024) { a2 -= (lockspeed<<1); if (a2 < 0) a2 = 0; }
				else           { a2 += (lockspeed<<1); if (a2 >= 2048) a2 = 0; }
			}
			if (a1 < 1024) { a1 -= (lockspeed<<1); if (a1 < 0) a1 = 0; }
			else           { a1 += (lockspeed<<1); if (a1 >= 2048) a1 = 0; }
		}

		a1 = ((a1+((a1vel*lockspeed)>>(5-shiftstate))+2048)&2047);
		a2 = ((a2+((a2vel*lockspeed)>>(5-shiftstate))+2048)&2047);
		a3 = ((a3+((a3vel*lockspeed)>>(5-shiftstate))+2048)&2047);
			//Auto-Tilting
		if (a3vel == 0)
		{
			if (a1 < 1024)
			{
				a1 -= (lockspeed<<1);
				if (a1 < 0) a1 = 0;
			}
			else
			{
				a1 += (lockspeed<<1);
				if (a1 >= 2048) a1 = 0;
			}
		}
		else
		{
			a1 = (((((((a1+1024)&2047)*7) + (1024-a3vel))>>3)+1024)&2047);
		}

		if (keystatus[0x26])     // L (load)
		{
			keystatus[0x26] = 0;
			loadboard("board.cub");
		}
		if (keystatus[0x1f]&(keystatus[0x38]|keystatus[0xb8]))     // Alt-S (save)
		{
			keystatus[0x1f] = 0;
			saveboard("board.cub");
		}

		if (keystatus[0xd2])  // Insert cube
		{
			keystatus[0xd2] = 0;
			searchit();
			x = ((searchcubak/(BSIZ*BSIZ))<<10)+512;
			y = (((searchcubak/BSIZ)%BSIZ)<<10)+512;
			z = ((searchcubak%BSIZ)<<10)+512;
			if ((klabs(posx-x) > 512+WALLDIST) ||
				 (klabs(posy-y) > 512+WALLDIST) ||
				 (klabs(posz-z) > 512+WALLDIST))
			{
				if (temppicnum >= 0) i = temppicnum; else i = 0;
				for(j=0;j<6;j++) board[j*BSIZ*BSIZ*BSIZ+searchcubak] = i;
			}
		}
		if (keystatus[0xd3])  // Delete cube
		{
			keystatus[0xd3] = 0;
			searchit();
			x = searchcube/(BSIZ*BSIZ);
			y = (searchcube/BSIZ)%BSIZ;
			z = searchcube%BSIZ;
			if ((x > 0) && (x < BSIZ-1))
				if ((y > 0) && (y < BSIZ-1))
					if ((z > 0) && (z < BSIZ-1))
						 board[searchcube] = -1;
		}
		if (keystatus[0xf])   // Tab
		{
			keystatus[0xf] = 0;
			searchit();
			temppicnum = board[searchcube+searchedge*BSIZ*BSIZ*BSIZ];
		}
		if (keystatus[0x1c])  // L. Enter
		{
			keystatus[0x1c] = 0;
			if (temppicnum >= 0)
			{
				searchit();
				board[searchcube+searchedge*BSIZ*BSIZ*BSIZ] = temppicnum;
			}
		}
		if (keystatus[0x2f])  // V
		{
			keystatus[0x2f] = 0;
			searchit();
			i = searchcube+searchedge*BSIZ*BSIZ*BSIZ;
			if (board[i] >= 0) board[i] = gettile(board[i]);
		}

		lockspeed = totalclock-baktotalclock;
		baktotalclock += lockspeed;
		numframes++;
	}
	_dos_setvect(0x9, oldkeyhandler);
	_dos_setvect(0x8, oldtimerhandler);
	koutp(0x43,54); koutp(0x40,255); koutp(0x40,255);
	unloadpics();
	uninitmouse();
	setvmode(0x3);
	if (totalclock != 0)
	{
		i = (numframes*12000L)/totalclock;
		printf("%d.%1d%1d frames per second\n",(short)(i/100),(short)((i/10)%10),(short)(i%10));
	}
	free(screen);
}

loadboard(char *filename)
{
	long fil, i, j;

	if (klabs(of[0]) <= 1) for(i=6*4*3-1;i>=0;i--) of[i] <<= 9;

	if ((fil = open(filename,O_BINARY|O_RDONLY,S_IREAD)) == -1)
	{
		for(i=6*BSIZ*BSIZ*BSIZ-1;i>=0;i--) board[i] = 0;
		posx = (BSIZ<<9)+512;
		posy = (BSIZ<<9)+512;
		posz = (BSIZ<<9)+512;
		board[(posx>>10)*BSIZ*BSIZ+(posy>>10)*BSIZ+(posz>>10)] = -1;
		a1 = 0; a2 = 0; a3 = 0;
		return;
	}
	read(fil,&i,4);
	if (i != 0xcbe)
	{
		printf("Wrong map version\n");
		exit(0);
	}
	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&a1,4);
	read(fil,&a2,4);
	read(fil,&a3,4);
	read(fil,board,6*BSIZ*BSIZ*BSIZ*2);
	close(fil);
}

saveboard(char *filename)
{
	long i, fil;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return;

	i = 0xcbe;
	write(fil,&i,4);
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&a1,4);
	write(fil,&a2,4);
	write(fil,&a3,4);
	write(fil,board,6*BSIZ*BSIZ*BSIZ*2);
	close(fil);
}

loadtables()
{
	long i, fil;

	for(i=0;i<2048;i++)
		sqrtable[i] = calcksqrt(i<<20);

	if ((fil = open("tables.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&sintable[0],4096);
		lseek(fil,4096+4096+640,SEEK_SET);
		read(fil,&textfont[0],1024);
		read(fil,&smalltextfont[0],1024);
		read(fil,&britable[0],1024);
		close(fil);
	}
	if ((fil = open("setup.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
	{
		read(fil,&option[0],NUMOPTIONS);
		read(fil,&keys[0],NUMKEYS);
		close(fil);
	}
}

loadpalette()
{
	long i, j, fil;

	if (paletteloaded == 0)
	{
		if ((fil = open("palette.dat",O_BINARY|O_RDWR,S_IREAD)) != -1)
		{
			numpalookups = ((filelength(fil)-768)>>7);
			if ((numpalookups&1) > 0)
			{   //Transluscent palette found in palette.dat
				numpalookups = ((numpalookups-255)>>1);

				if (translucloaded == 0)
					if ((transluc = (char *)malloc(65536L)) != 0)
					{
						//fixtransluscence(FP_OFF(transluc));
						translucloaded = 1;
					}
			}
			else
			{    //No transluscent palette in palette.dat
				numpalookups >>= 1;
			}
			read(fil,&palette[0],768);
			read(fil,&palookup[0][0],(numpalookups<<8));
			if (translucloaded == 1)
			{
				for(i=0;i<255;i++)
				{
					read(fil,&transluc[(i<<8)+i+1],255-i);
					for(j=i+1;j<256;j++)
						transluc[(j<<8)+i] = transluc[(i<<8)+j];
				}
				for(i=0;i<256;i++)
					transluc[(i<<8)+i] = i;
			}

			blackcol = 0; j = 0x7fffffff;
			for(i=0;i<256;i++)
				if (palette[i*3]+palette[i*3+1]+palette[i*3+2] < j)
				{
					j = palette[i*3]+palette[i*3+1]+palette[i*3+2];
					blackcol = i;
				}

			whitecol = 0; j = 0x80000000;
			for(i=0;i<256;i++)
				if (palette[i*3]+palette[i*3+1]+palette[i*3+2] > j)
				{
					j = palette[i*3]+palette[i*3+1]+palette[i*3+2];
					whitecol = i;
				}

			close(fil);
		}
		paletteloaded = 1;
	}
}

drawscreen()
{
	long i;
	char tempbuf[256];

	if (keystatus[0x38]) clearbuf(frameplace,(xdim*ydim)>>2,0L);

	drawcubes();

	i = totalclock;
	if (i != frameval[framecnt])
	{
		sprintf(tempbuf,"%ld",(120*AVERAGEFRAMES)/(i-frameval[framecnt]));
		printext256(0L,0L,31,-1,tempbuf,1);
		frameval[framecnt] = i;
	}
	framecnt = ((framecnt+1)&(AVERAGEFRAMES-1));

	sprintf(tempbuf,"Regions: %ld",globalregioncnt);
	printext256(148L,0L,31,-1,tempbuf,1);
	sprintf(tempbuf,"Splits: %ld",globalsplitcnt);
	printext256(148L,6L,31,-1,tempbuf,1);

	obstatus = bstatus, getmousevalues(&mousx,&mousy,&bstatus);
	searchx += mousx;
	searchy += mousy;
	if (searchx < 4) searchx = 4;
	if (searchx >= xdim-4) searchx = xdim-4-1;
	if (searchy < 4) searchy = 4;
	if (searchy >= ydim-4) searchy = ydim-4-1;

	for(i=1;i<=4;i++)
	{
		drawpixel(frameplace+ylookup[searchy+i]+(searchx+0),whitecol);
		drawpixel(frameplace+ylookup[searchy-i]+(searchx+0),whitecol);
		drawpixel(frameplace+ylookup[searchy+0]+(searchx+i),whitecol);
		drawpixel(frameplace+ylookup[searchy+0]+(searchx-i),whitecol);
	}

	nextpage();
}

nextpage()
{
	long i, totbytes;

	switch(option[0])
	{
		case 0: case 1:
			totbytes = xdim*ydim;
			for(i=0;i<totbytes;i+=65536)
			{
				if ((xdim > 320) || (ydim > 200)) vesasetpage(i>>16);
				copybuf(frameplace+i,0xa0000,min(totbytes-i,65536)>>2);
			}
			break;
		case 2:
			koutp(0x3d4,0xc); koutp(0x3d5,visualpage<<6);
			visualpage = ((visualpage+1)&3);
			koutp(0x3cd,visualpage+(visualpage<<4));
			break;
		case 3:
			koutp(0x3d4,0xc); koutp(0x3d5,visualpage<<6);
			visualpage = ((visualpage+1)&3);
			koutp(0x3ce,0x9); koutp(0x3cf,visualpage<<4);
			break;
		case 4:
			koutp(0x3d4,0xc); koutp(0x3d5,visualpage<<6);
			visualpage = ((visualpage+1)&3);
			koutp(0x3d4,0x35); koutp(0x3d5,visualpage);
			break;
	}
}

drawcubes()
{
	long i, j, k, x, y, z, xx, yy, zz, p, *optr, firstime;
	long cubelistplc, cubelistnum, myindex, daindex, dacube, dapicnum;

	initpoly();

	x = max(min(posx>>10,BSIZ-1),0);
	y = max(min(posy>>10,BSIZ-1),0);
	z = max(min(posz>>10,BSIZ-1),0);
	cubelist[0] = x*BSIZ*BSIZ+y*BSIZ+z; cubelistplc = 0; cubelistnum = 1;
	do
	{
		myindex = cubelist[cubelistplc++];
		x = ((myindex/(BSIZ*BSIZ))<<10)+512-posx;
		y = (((myindex/BSIZ)%BSIZ)<<10)+512-posy;
		z = ((myindex%BSIZ)<<10)+512-posz;

		for(j=0;j<=5;j++)
		{
			switch(j)
			{
				case 0: if (x < -512) continue; daindex = myindex+BSIZ*BSIZ; break;
				case 1: if (y < -512) continue; daindex = myindex+BSIZ; break;
				case 2: if (z < -512) continue; daindex = myindex+1; break;
				case 3: if (x > 512) continue; daindex = myindex-BSIZ*BSIZ; break;
				case 4: if (y > 512) continue; daindex = myindex-BSIZ; break;
				case 5: if (z > 512) continue; daindex = myindex-1; break;
			}
			dacube = board[daindex];
			if (dacube >= 0) dacube = board[daindex+j*BSIZ*BSIZ*BSIZ];

			if (dacube < 0)
			{
				for(i=0;i<cubelistnum;i++)     //0 was cubelistplc
					if (cubelist[i] == daindex) break;
				if (i < cubelistnum) continue;
			}

			optr = (long *)(of+j*12);
			for(i=3;i>=0;i--)
			{
				r[i].x = ((optr[0]+x)<<16);
				r[i].y = ((optr[1]+y)<<16);
				r[i].z = ((optr[2]+z)<<16);
				optr += 3;
			}
			point2[0] = 1; point2[1] = 2; point2[2] = 3; point2[3] = 0;
			poly[0] = 0; poly[1] = 4; numpolys = 1; numpoints = 4;

			if (dacube < 0)
			{
				xx = ((r[0].x+r[1].x+r[2].x+r[3].x)>>2);
				yy = ((r[0].y+r[1].y+r[2].y+r[3].y)>>2);
				zz = ((r[0].z+r[1].z+r[2].z+r[3].z)>>2);

				i = 0;
				do
				{
					if (tmulscale32(bn[i].x,xx,bn[i].y,yy,bn[i].z,zz) < 0)
						i = front[i];
					else
						i = back[i];
				} while (i >= 0);

				if (i == -2) { cubelist[cubelistnum++] = daindex; continue; }
			}

			addpoly();

			if (dacube < 0)
			{
				for(p=numpolys-1;p>=0;p--)
					if ((polytag[p] == -2) && (poly[p+1] >= poly[p]+3))
						{ cubelist[cubelistnum++] = daindex; break; }
				continue;
			}

			firstime = 0;
			for(p=numpolys-1;p>=0;p--)
				if ((polytag[p] == -2) && (poly[p+1] >= poly[p]+3))
				{
					for(i=poly[p];i<poly[p+1];i++)
					{
						xx = r[i].x; yy = r[i].y; zz = r[i].z;
						rotate3d(xx,yy,zz);

						sc[i].x = scale(xreszoom,xx,zz) + ((xdim+1)<<15);
						sc[i].y = scale(xreszoom,yy,zz) + ((ydim+1)<<15);

						if (sc[i].x < 0) sc[i].x = 0;
						if (sc[i].x > (xdim<<16)) sc[i].x = (xdim<<16);
						if (sc[i].y < 0) sc[i].y = 0;
						if (sc[i].y > (ydim<<16)) sc[i].y = (ydim<<16);
					}

					k = 0;
					for(i=poly[p];i<poly[p+1];i++)
						k += mulscale28(sc[i].x+sc[point2[i]].x,sc[point2[i]].y-sc[i].y);

					if (k > 0)
					{
						globalregioncnt++;

						if (firstime == 0)
						{
							firstime = 1;
							gtext = dacube;
							if (j >= 3) gshade = ((j-3)<<2); else gshade = (j<<2);
							initfasttextureplane();
						}
						drawfasttextureplane((long)poly[p],(long)poly[p+1]-1);
						arealeft -= k; if (arealeft <= 0) return;

						back[bak[p]] = add2bsp(p);  //If == -1, then deadbranch!
					}
				}
		}
	} while (cubelistplc < cubelistnum);
}

setuprotate3d()
{
	long c1, s1, c2, s2, c3, s3, c1c3, c1s3, s1c3, s1s3;

	c1 = sintable[(a1+512)&2047]; s1 = sintable[a1&2047];
	c2 = sintable[(a2+512)&2047]; s2 = sintable[a2&2047];
	c3 = sintable[(a3+512)&2047]; s3 = sintable[a3&2047];

	c1c3 = c1*c3; c1s3 = c1*s3; s1c3 = s1*c3; s1s3 = s1*s3;

	gxx = mulscale14(s1s3,s2) + c1c3;
	gxy = -mulscale14(c1s3,s2) + s1c3;
	gxz = c2*s3;
	gyx = -s1*c2;
	gyy = c1*c2;
	gyz = (s2<<14);
	gzx = mulscale14(s1c3,s2) - c1s3;
	gzy = -mulscale14(c1c3,s2) - s1s3;
	gzz = c2*c3;
}

ksqrt(long num)
{
	return(ksqrtasm(num));
}

calcksqrt(long num)
{
	long root, temp;

	if (num < 0) return(-1);
	root = 128;
	do
	{
		temp = root;
		root = ((root+(num/root))>>1);
	}
	while (klabs(temp-root) > 1);
	return(root);
}

void initfasttextureplane()
{
	long i;

	gtext += animateoffs(gtext);
	if (waloff[gtext] == 0) loadtile((short)gtext);
	if (waloff[gtext] == 0) return;
	gbuf = (char *)(waloff[gtext]);

	i = picsiz[gtext];
	setupdline(waloff[gtext],FP_OFF(palookup[0]),(256<<(i&15))+(1<<(i>>4))-257);
	gmask1 = (((1<<(i&15))-1)<<8); gmask2 = (1<<(i>>4))-1;

	setupgetreadpos();
	if (klabs(gdx) <= klabs(gdy))
	{
		if (gdx == 0)
			gm = 0;
		else
		{
			gm = divscale16(gdx,gdy);   //d& = y& + mulscale&(x&, m&, 16)
			if (klabs(gm) < 32) gm = 0;
			if (gm > 65536-64) gm = 65536;
			if (gm < -65536+64) gm = -65536;
		}

		if (gm > 0) hplc[0] = 0L; else hplc[0] = xdim;

		gxoffs = gux-mulscale16(guy,gm);
		gyoffs = gvx-mulscale16(gvy,gm);

		pinc[0] = 1L;
		if (gm < 0) pinc[2] = 1+xdim; else pinc[2] = 1-xdim;
	}
	else
	{
		if (gdy == 0)
			gm = 0;
		else
		{
			gm = divscale16(gdy,gdx);
			if (klabs(gm) < 32) gm = 0;
			if (gm > 65536-64) gm = 65536;
			if (gm < -65536+64) gm = -65536;
		}

		if (gm > 0) hplc[0] = 0L; else hplc[0] = xdim;

		gxoffs = guy-mulscale16(gux,gm);
		gyoffs = gvy-mulscale16(gvx,gm);

		pinc[0] = xdim;
		if (gm < 0) pinc[2] = xdim+1L; else pinc[2] = xdim-1L;
	}
}

setupgetreadpos()
{
	long i, x0, y0, z0, x1, y1, z1, x2, y2, z2; z0 = (r[0].z>>6);

	x0 = (r[0].x>>6); y0 = (r[0].y>>6);

	i = picsiz[gtext];
	if (((1<<(i&15)) == tilesizx[gtext]) && ((1<<(i>>4)) == tilesizy[gtext]) && ((tilesizx[gtext] < 64) || (tilesizy[gtext] < 64)))
	{
		x1 = ((r[1].x-r[0].x)>>6);
		y1 = ((r[1].y-r[0].y)>>6);
		z1 = ((r[1].z-r[0].z)>>6);
		x2 = ((r[3].x-r[0].x)>>6);
		y2 = ((r[3].y-r[0].y)>>6);
		z2 = ((r[3].z-r[0].z)>>6);
	}
	else
	{
		x1 = (r[1].x-r[0].x)/(tilesizx[gtext]);
		y1 = (r[1].y-r[0].y)/(tilesizx[gtext]);
		z1 = (r[1].z-r[0].z)/(tilesizx[gtext]);
		x2 = (r[3].x-r[0].x)/(tilesizy[gtext]);
		y2 = (r[3].y-r[0].y)/(tilesizy[gtext]);
		z2 = (r[3].z-r[0].z)/(tilesizy[gtext]);
	}

	rotate3d(x0,y0,z0);
	rotate3d(x1,y1,z1);
	rotate3d(x2,y2,z2);
	gdx = dmulscale24(y1,z2,-y2,z1); gdy = dmulscale24(x2,z1,-x1,z2);
	gux = dmulscale24(y2,z0,-y0,z2); guy = dmulscale24(x0,z2,-x2,z0);
	gvx = dmulscale24(y0,z1,-y1,z0); gvy = dmulscale24(x1,z0,-x0,z1);
	gd = mulscale16(dmulscale24(x1,y2,-x2,y1),xreszoom)-((gdx*(xdim-1)+gdy*(ydim-1))>>1);
	gu = mulscale16(dmulscale24(x2,y0,-x0,y2),xreszoom)-((gux*(xdim-1)+guy*(ydim-1))>>1);
	gv = mulscale16(dmulscale24(x0,y1,-x1,y0),xreszoom)-((gvx*(xdim-1)+gvy*(ydim-1))>>1);
	gt = mulscale28(tmulscale16(x0,gdx,x1,gux,x2,gvx),xreszoom);
}

void drawfasttextureplane(long pstart, long pend)
{
	long i, j, z, zz, x, y, x1, y1, x2, y2, d1, d2, dapicnum;
	long plc, inc, minx, maxx, miny, maxy, c1, c2;

	if (keystatus[0x38])
	{
		for(z=pstart;z<=pend;z++)
		{
			zz = point2[z];
			drawline256(sc[z].x,sc[z].y,sc[zz].x,sc[zz].y,15);
		}
		return;
	}

	if (gbuf == NULL) return;

	if (klabs(gdx) <= klabs(gdy))
	{
		for(z=pstart;z<=pend;z++) sc[z].y -= 65536;

		minx = sc[pstart].x; maxx = sc[pstart].x;
		for(z=pstart+1;z<=pend;z++)
		{
			x = sc[z].x;
			if (x < minx) minx = x;
			if (x > maxx) maxx = x;
		}
		minx >>= 16; maxx >>= 16;

		qinterpolatedown16(&hplc[minx],maxx-minx+1,(((long)hplc[0])<<16)+gm*minx+32768,gm);

		if (gm != 0)
		{
			if (gm < 0) i = xdim; else i = -xdim;
			for(x=minx;x<=maxx;x++)
			{
				hinc[x] = 1; hind[x] = 0;
				if (hplc[x] != hplc[x+1]) hinc[x] += i, hind[x] = 8;
			}
			hinc[maxx] = 0;
		}
		else
		{
			for(x=minx;x<=maxx;x++)
				hinc[x] = 1, hind[x] = 0;
		}

		miny = 0x7fffffff; maxy = 0x80000000;
		for(z=pstart;z<=pend;z++)
		{
			y = (sc[z].y>>16) + hplc[sc[z].x>>16];
			if (y < miny) miny = y;
			if (y > maxy) maxy = y;
		}
		miny--; maxy++;

		c1 = 1; clearbuf(&h1[miny],((maxy+1-miny)+1)>>1,0L);
		c2 = 1; clearbuf(&h2[miny],((maxy+1-miny)+1)>>1,0L);

		for(z=pstart;z<=pend;z++)
		{
			zz = point2[z];
			x1 = sc[z].x; y1 = sc[z].y; x2 = sc[zz].x; y2 = sc[zz].y;

			if (((x1^x2)&0xffff0000) == 0)
			{
				x = (x1>>16);
				y = (y1>>16) + hplc[x];
				j = (y2>>16) + hplc[x];
				while (j < y) { p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; y--; }
				while (j > y) { y++; p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; }
			}
			else
			{
				if ((klabs(y2-y1)>>15) < klabs(x2-x1))
					inc = divscale16(y2-y1,x2-x1);
				else
					inc = 0;

				d1 = (x1>>16); d2 = (x2>>16);
				if (x1 < x2)
				{
					plc = y1 + mulscale16((-x1)&65535,inc);
					y = (y1>>16) + hplc[d1];
					for(x=d1;x<d2;x++)
					{
						j = (plc>>16) + hplc[x]; plc += inc;
						while (j < y) { p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; y--; }
						while (j > y) { y++; p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; }
					}
					j = (y2>>16) + hplc[d2];
					while (j < y) { p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; y--; }
					while (j > y) { y++; p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; }
				}
				else
				{
					plc = y2 + mulscale16((-x2)&65535,inc);
					y = (y2>>16) + hplc[d2];
					for(x=d2;x<d1;x++)
					{
						j = (plc>>16) + hplc[x]; plc += inc;
						while (j < y) { p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; y--; }
						while (j > y) { y++; p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; }
					}
					j = (y1>>16) + hplc[d1];
					while (j < y) { p2[c2] = x; n2[c2] = h2[y]; h2[y] = c2; c2++; y--; }
					while (j > y) { y++; p1[c1] = x; n1[c1] = h1[y]; h1[y] = c1; c1++; }
				}
			}
		}

		for(y=miny;y<=maxy;y++)
			for(y1=h1[y],y2=h2[y];y1!=0;y1=n1[y1],y2=n2[y2])
			{
				d1 = y1; d2 = y2;
				for(i=n1[y1],j=n2[y2];i!=0;i=n1[i],j=n2[j])
				{
					if (p1[i] < p1[d1]) d1 = i;
					if (p2[j] < p2[d2]) d2 = j;
				}
				if (p1[d1] < p2[d2]) diaghline((long)p1[d1],(long)p2[d2]-1,y);
				p1[d1] = p1[y1]; p2[d2] = p2[y2];
			}
	}
	else
	{
		for(z=pstart;z<=pend;z++) sc[z].x -= 65536;

		miny = sc[pstart].y; maxy = sc[pstart].y;
		for(z=pstart+1;z<=pend;z++)
		{
			y = sc[z].y;
			if (y < miny) miny = y;
			if (y > maxy) maxy = y;
		}
		miny >>= 16; maxy >>= 16;

		qinterpolatedown16(&hplc[miny],maxy-miny+1,(((long)hplc[0])<<16)+gm*miny+32768,gm);

		if (gm != 0)
		{
			if (gm < 0) i = 1; else i = -1;
			for(y=miny;y<maxy;y++)
			{
				hinc[y] = xdim, hind[y] = 0;
				if (hplc[y] != hplc[y+1]) hinc[y] += i, hind[y] = 8;
			}
			hinc[maxy] = 0;
		}
		else
		{
			for(y=miny;y<=maxy;y++)
				hinc[y] = xdim, hind[y] = 0;
		}

		minx = 0x7fffffff; maxx = 0x80000000;
		for(z=pstart;z<=pend;z++)
		{
			x = (sc[z].x>>16) + hplc[sc[z].y>>16];
			if (x < minx) minx = x;
			if (x > maxx) maxx = x;
		}
		minx--; maxx++;

		c1 = 1; clearbuf(&h1[minx],((maxx+1-minx)+1)>>1,0L);
		c2 = 1; clearbuf(&h2[minx],((maxx+1-minx)+1)>>1,0L);

		for(z=pstart;z<=pend;z++)
		{
			zz = point2[z];
			x1 = sc[z].x; y1 = sc[z].y; x2 = sc[zz].x; y2 = sc[zz].y;

			if (((y1^y2)&0xffff0000) == 0)
			{
				y = (y1>>16);
				x = (x1>>16) + hplc[y];
				j = (x2>>16) + hplc[y];
				while (j < x) { p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; x--; }
				while (j > x) { x++; p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; }
			}
			else
			{
				if ((klabs(x2-x1)>>15) < klabs(y2-y1))
					inc = divscale16(x2-x1,y2-y1);
				else
					inc = 0;

				d1 = (y1>>16); d2 = (y2>>16);
				if (y1 < y2)
				{
					plc = x1 + mulscale16((-y1)&65535,inc);
					x = (x1>>16) + hplc[d1];
					for(y=d1;y<d2;y++)
					{
						j = (plc>>16) + hplc[y]; plc += inc;
						while (j < x) { p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; x--; }
						while (j > x) { x++; p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; }
					}
					j = (x2>>16) + hplc[d2];
					while (j < x) { p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; x--; }
					while (j > x) { x++; p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; }
				}
				else
				{
					plc = x2 + mulscale16((-y2)&65535,inc);
					x = (x2>>16) + hplc[d2];
					for(y=d2;y<d1;y++)
					{
						j = (plc>>16) + hplc[y]; plc += inc;
						while (j < x) { p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; x--; }
						while (j > x) { x++; p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; }
					}
					j = (x1>>16) + hplc[d1];
					while (j < x) { p2[c2] = y; n2[c2] = h2[x]; h2[x] = c2; c2++; x--; }
					while (j > x) { x++; p1[c1] = y; n1[c1] = h1[x]; h1[x] = c1; c1++; }
				}
			}
		}

		for(x=minx;x<=maxx;x++)
			for(x1=h1[x],x2=h2[x];x1!=0;x1=n1[x1],x2=n2[x2])
			{
				d1 = x1; d2 = x2;
				for(i=n1[x1],j=n2[x2];i!=0;i=n1[i],j=n2[j])
				{
					if (p1[i] < p1[d1]) d1 = i;
					if (p2[j] < p2[d2]) d2 = j;
				}
				if (p2[d2] < p1[d1]) diagvline((long)p2[d2],(long)p1[d1]-1,x);
				p1[d1] = p1[x1]; p2[d2] = p2[x2];
			}
	}
}

diaghline (long sx, long end, long sy)
{
	long p, d, d2, shade, u, v, uu[2], vv[2];
	char *dapal;

	d = gd + (sy-hplc[0])*gdy; if (d <= 512) return; d = divscale22(65536,d);

	d2 = mulscale29(gt,d);
	shade = max(min(gshade+d2+((keystatus[0x20]<<2)&sy),numpalookups-1),0);
	sy -= hplc[sx]; p = ylookup[sy]+sx+frameplace;
	if (sx >= end)
	{
		u = mulscale8(gu+sx*gux+sy*guy,d);
		v = mulscale8(gv+sx*gvx+sy*gvy,d);
		dapal = (char *)((shade<<8)+FP_OFF(palookup[0]));
		drawpixel(p,dapal[gbuf[((u>>16)&gmask1)+((v>>24)&gmask2)]]);
		return;
	}

	if (((gm == 0) || (d2 == 0) || (keystatus[0xb8] != 0)) && (keystatus[0x36] == 0))
	{
		sy = ((sy+hplc[sx]-hplc[0])<<16)-sx*gm;
		u = mulscale8(gu+sx*gux+mulscale16(sy,guy),d);
		v = mulscale8(gv+sx*gvx+mulscale16(sy,gvy),d);
		uu[0] = uu[1] = mulscale8(gxoffs,d);
		vv[0] = vv[1] = mulscale8(gyoffs,d);
	}
	else
	{
		u = mulscale8(gu+sx*gux+sy*guy,d);
		v = mulscale8(gv+sx*gvx+sy*gvy,d);
		uu[0] = uu[1] = mulscale8(gux,d);
		vv[0] = vv[1] = mulscale8(gvx,d);
		if (gm < 0) { uu[1] += mulscale8(guy,d); vv[1] += mulscale8(gvy,d); }
				 else { uu[1] -= mulscale8(guy,d); vv[1] -= mulscale8(gvy,d); }
	}
	ui[0] = (uu[0]<<8); ui[1] = (uu[0]>>24);
	ui[2] = (uu[1]<<8); ui[3] = (uu[1]>>24);
	vi[0] = (vv[0]<<8); vi[1] = (vv[0]>>24);
	vi[2] = (vv[1]<<8); vi[3] = (vv[1]>>24);
	dline(end,u,v,shade,sx,p);
}

diagvline (long sy, long end, long sx)
{
	long p, d, d2, shade, u, v, uu[2], vv[2];
	char *dapal;

	d = gd + (sx-hplc[0])*gdx; if (d <= 512) return; d = divscale22(65536,d);

	d2 = mulscale29(gt,d);
	shade = max(min(gshade+d2+((keystatus[0x20]<<2)&sx),numpalookups-1),0);
	sx -= hplc[sy]; p = ylookup[sy]+sx+frameplace;
	if (sy >= end)
	{
		u = mulscale8(gu+sx*gux+sy*guy,d);
		v = mulscale8(gv+sx*gvx+sy*gvy,d);
		dapal = (char *)((shade<<8)+FP_OFF(palookup[0]));
		drawpixel(p,dapal[gbuf[((u>>16)&gmask1)+((v>>24)&gmask2)]]);
		return;
	}

	if (((gm == 0) || (d2 == 0) || (keystatus[0xb8] != 0)) && (keystatus[0x36] == 0))
	{
		sx = ((sx+hplc[sy]-hplc[0])<<16)-sy*gm;
		u = mulscale8(gu+mulscale16(sx,gux)+sy*guy,d);
		v = mulscale8(gv+mulscale16(sx,gvx)+sy*gvy,d);
		uu[0] = uu[1] = mulscale8(gxoffs,d);
		vv[0] = vv[1] = mulscale8(gyoffs,d);
	}
	else
	{
		u = mulscale8(gu+sx*gux+sy*guy,d);
		v = mulscale8(gv+sx*gvx+sy*gvy,d);
		uu[0] = uu[1] = mulscale8(guy,d);
		vv[0] = vv[1] = mulscale8(gvy,d);
		if (gm < 0) { uu[1] += mulscale8(gux,d); vv[1] += mulscale8(gvx,d); }
				 else { uu[1] -= mulscale8(gux,d); vv[1] -= mulscale8(gvx,d); }
	}
	ui[0] = (uu[0]<<8); ui[1] = (uu[0]>>24);
	ui[2] = (uu[1]<<8); ui[3] = (uu[1]>>24);
	vi[0] = (vv[0]<<8); vi[1] = (vv[0]>>24);
	vi[2] = (vv[1]<<8); vi[3] = (vv[1]>>24);
	dline(end,u,v,shade,sy,p);
}

initpoly()
{
	long i;

	arealeft = mulscale(xdim*ydim,4095,7);

	pnum = 4;
	bn[0].x = -65536; bn[0].y = 0; bn[0].z = 65536;
	bn[1].x = 65536; bn[1].y = 0; bn[1].z = 65536;
	bn[2].x = 0; bn[2].y = -65536; bn[2].z = ydimscale;
	bn[3].x = 0; bn[3].y = 65536; bn[3].z = ydimscale;

	for(i=0;i<pnum;i++)
	{
		backrotate3d(bn[i].x,bn[i].y,bn[i].z);
		front[i] = -1; back[i] = i+1;
	}
	back[pnum-1] = -2;

	globalregioncnt = 0;
	globalsplitcnt = 0;
}

addpoly()
{
	long i, j, p, pcnt, x, y, z, x0, x1, y0, y1, z0, z1;

	for(i=numpoints-1;i>=0;i--) splittag[i] = (i|0x8000);

			//Get bounding box
	x0 = y0 = z0 = 0x7fffffff;
	x1 = y1 = z1 = 0x80000000;
	for(i=numpoints-1;i>=0;i--)
	{
		if (r[i].x < x0) x0 = r[i].x;
		if (r[i].x > x1) x1 = r[i].x;
		if (r[i].y < y0) y0 = r[i].y;
		if (r[i].y > y1) y1 = r[i].y;
		if (r[i].z < z0) z0 = r[i].z;
		if (r[i].z > z1) z1 = r[i].z;
	}

	polytag[0] = 0; pcnt = 0;
	do
	{
		p = polytag[pcnt];
		if ((p >= 0) && (poly[pcnt+1] >= poly[pcnt]+3))
		{
				//Oops!  Dead branch
			if ((front[p] < 0) && (back[p] == -1)) { pcnt++; continue; }

			/*   //Get bounding box
			j = poly[pcnt];
			x0 = r[j].x; x1 = r[j].x;
			y0 = r[j].y; y1 = r[j].y;
			z0 = r[j].z; z1 = r[j].z;
			for(i=poly[pcnt+1]-1;i>j;i--)
			{
				if (r[i].x < x0) x0 = r[i].x;
				if (r[i].x > x1) x1 = r[i].x;
				if (r[i].y < y0) y0 = r[i].y;
				if (r[i].y > y1) y1 = r[i].y;
				if (r[i].z < z0) z0 = r[i].z;
				if (r[i].z > z1) z1 = r[i].z;
			} */

				//Bounding box code
			i = 0; j = 0;
			x = bn[p].x; y = bn[p].y; z = bn[p].z;
			if (x > 0) i += mulscale22(x,x1), j += mulscale22(x,x0);
			if (x < 0) i += mulscale22(x,x0), j += mulscale22(x,x1);
			if (y > 0) i += mulscale22(y,y1), j += mulscale22(y,y0);
			if (y < 0) i += mulscale22(y,y0), j += mulscale22(y,y1);
			if (z > 0) i += mulscale22(z,z1), j += mulscale22(z,z0);
			if (z < 0) i += mulscale22(z,z0), j += mulscale22(z,z1);
			if (i <= 0) { polytag[pcnt] = front[p]; bak[pcnt] = p; continue; }
			if (j >= 0) { polytag[pcnt] = back[p]; bak[pcnt] = p; continue; }

				//Bounding box fails
			if (front[p] < 0)
			{
				polytag[numpolys] = back[p]; bak[numpolys] = p;
				splitpoly1(p,pcnt,-x,-y,-z);
			}
			else if (back[p] == -1)    //Already filled region
			{
				polytag[numpolys] = front[p]; bak[numpolys] = p;
				splitpoly1(p,pcnt,x,y,z);
			}
			else
			{
				polytag[numpolys+0] = front[p]; bak[numpolys+0] = p;
				polytag[numpolys+1] = back[p]; bak[numpolys+1] = p;
				splitpoly2(p,pcnt,x,y,z);
			}
		}
		pcnt++;
	} while (pcnt < numpolys);
}

add2bsp (long polynum)
{
	long z, zz, dap, i, pcnt, plistnum, daparent;
	long bx1, by1, bz1, bx2, by2, bz2, t1, t2, fz, bz;

	plistnum = 0;
	for(z=poly[polynum];z<poly[polynum+1];z++)
		if (splittag[z] < 0) plist[plistnum++] = (z&0x7fff);
	if (plistnum == 0) return(-1);

	daparent = pnum; pcnt = pnum+1;


	for(dap=plistnum-1;dap>=0;dap--)
	{
		i = mulscale16(krand(),dap+1);
		z = plist[i];  bx1 = r[z].x; by1 = r[z].y; bz1 = r[z].z;
		z = point2[z]; bx2 = r[z].x; by2 = r[z].y; bz2 = r[z].z;
		plist[i] = plist[dap];

		bn[pnum].x = -dmulscale26(by1,bz2,-by2,bz1);
		bn[pnum].y = -dmulscale26(bx2,bz1,-bx1,bz2);
		bn[pnum].z = -dmulscale26(bx1,by2,-bx2,by1);
		reduce(bn[pnum].x,bn[pnum].y,bn[pnum].z,32768);

		if (dap == 0) front[pnum] = -1; else front[pnum] = pnum+1;
		back[pnum++] = -2;
	}
	return(daparent);

	/*
		//GOOD IDEA: COMPARE SIGNS WITH BOUNDING BOX CODE
	for(dap=plistnum-1;dap>=0;dap--)
	{
		i = 0; //mulscale16(krand(),dap+1);
		z = plist[i];  bx1 = r[z].x; by1 = r[z].y; bz1 = r[z].z;
		z = point2[z]; bx2 = r[z].x; by2 = r[z].y; bz2 = r[z].z;
		plist[i] = plist[dap];

		bn[pnum].x = -dmulscale26(by1,bz2,-by2,bz1);
		bn[pnum].y = -dmulscale26(bx2,bz1,-bx1,bz2);
		bn[pnum].z = -dmulscale26(bx1,by2,-bx2,by1);
		reduce(bn[pnum].x,bn[pnum].y,bn[pnum].z,32768);

		front[pnum] = -1; back[pnum] = -2; pt[pnum] = daparent;
		pnum++;

		while (pcnt < pnum)
		{
			z = pt[pcnt];
			do
			{
				t1 = tmulscale22(bn[z].x,bx1,bn[z].y,by1,bn[z].z,bz1);
				t2 = tmulscale22(bn[z].x,bx2,bn[z].y,by2,bn[z].z,bz2);

				if ((t1 <= 0) && (t2 <= 0))
				{
					if (front[z] >= 0) z = front[z];
					else { front[z] = pcnt; break; }
				}
				else if ((t1 >= 0) && (t2 >= 0))
				{
					if (back[z] >= 0) z = back[z];
					else { back[z] = pcnt; break; }
				}
				else
				{
					copybuf(&bn[pcnt],&bn[pnum],3);
					front[pnum] = -1; back[pnum] = -2;

					fz = front[z]; bz = back[z];
					if (t1 < 0)
					{
						if (bz >= 0) pt[pnum] = bz; else { back[z] = pnum; pt[pnum] = -1; }
						pnum++;
						if (fz >= 0) z = fz; else { front[z] = pcnt; break; }
					}
					else
					{
						if (fz >= 0) pt[pnum] = fz; else { front[z] = pnum; pt[pnum] = -1; }
						pnum++;
						if (bz >= 0) z = bz; else { back[z] = pcnt; break; }
					}
				}
			} while (z >= 0);
			pcnt++;
		}
	}
	return(daparent); */
}

splitpoly1 (long p, long polynum, long dx, long dy, long dz)
{
	long i, z, zz, zsgn, zzsgn, splitcnt, start2;
	long pstart, pend, s1, s2, s3, s4;

	globalsplitcnt++;
	pstart = poly[polynum]; pend = poly[polynum+1]-1;

	start2 = numpoints;
	z = pstart; splitcnt = 0;
	do
	{
		zzsgn = 0;
		if (dx != 0) zzsgn += mulscale22(r[z].x,dx);
		if (dy != 0) zzsgn += mulscale22(r[z].y,dy);
		if (dz != 0) zzsgn += mulscale22(r[z].z,dz);
		do
		{
			zz = point2[z]; point2[z] |= 0x8000;

			zsgn = zzsgn;
			zzsgn = 0;
			if (dx != 0) zzsgn += mulscale22(r[zz].x,dx);
			if (dy != 0) zzsgn += mulscale22(r[zz].y,dy);
			if (dz != 0) zzsgn += mulscale22(r[zz].z,dz);
			if (zsgn < 0)
			{
				copybuf(&r[z],&r[numpoints],3);
				splittag[numpoints] = splittag[z];
				point2[numpoints] = numpoints+1;
				numpoints++;
			}
			if ((zsgn^zzsgn) < 0)
			{
				i = divscale28(zsgn,zsgn-zzsgn);
				r[numpoints].x = r[z].x + mulscale28(r[zz].x-r[z].x,i);
				r[numpoints].y = r[z].y + mulscale28(r[zz].y-r[z].y,i);
				r[numpoints].z = r[z].z + mulscale28(r[zz].z-r[z].z,i);
				point2[numpoints] = numpoints+1;
				if (zsgn < 0)
					{ split[splitcnt++] = numpoints; splittag[numpoints] = p; }
				else
					splittag[numpoints] = splittag[z];
				numpoints++;
			}
			z = zz;
		} while (point2[z] >= 0);

		if (numpoints >= start2+3)
			{ point2[numpoints-1] = start2; start2 = numpoints; }
		else
			numpoints = start2;

		z = pend;
		while ((z > pstart) && (point2[z] < 0)) z--;
	} while (z > pstart);

	for(z=pstart;z<=pend;z++) point2[z] &= 0x7fff;

	/*
	dx = klabs(dx); dy = klabs(dy); dz = klabs(dz);
	if ((dx >= dy) && (dx >= dz))
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(r[s1].x-r[s4].x)+klabs(r[s3].x-r[s2].x) < klabs(r[s1].x-r[s2].x) + klabs(r[s3].x-r[s4].x))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}
	else if (dy >= dz)
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(r[s1].y-r[s4].y)+klabs(r[s3].y-r[s2].y) < klabs(r[s1].y-r[s2].y) + klabs(r[s3].y-r[s4].y))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}
	else
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(r[s1].z-r[s4].z)+klabs(r[s3].z-r[s2].z) < klabs(r[s1].z-r[s2].z) + klabs(r[s3].z-r[s4].z))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}*/

	numpolys++; poly[numpolys] = numpoints;
}

splitpoly2 (long p, long polynum, long dx, long dy, long dz)
{
	long i, z, zz, zsgn, zzsgn, splitcnt, splitcnt2, intx, inty, intz;
	long pstart, pend, numpoints2, start2, start22, s1, s2, s3, s4;

	globalsplitcnt++;
	pstart = poly[polynum]; pend = poly[polynum+1]-1;

	numpoints2 = MAXPOINTS-1;
	start2 = numpoints; start22 = numpoints2;
	z = pstart; splitcnt = 0; splitcnt2 = MAXSPLITCNT-1;
	do
	{
		zzsgn = 0;
		if (dx != 0) zzsgn += mulscale22(r[z].x,dx);
		if (dy != 0) zzsgn += mulscale22(r[z].y,dy);
		if (dz != 0) zzsgn += mulscale22(r[z].z,dz);
		do
		{
			zz = point2[z]; point2[z] |= 0x8000;

			zsgn = zzsgn;
			zzsgn = 0;
			if (dx != 0) zzsgn += mulscale22(r[zz].x,dx);
			if (dy != 0) zzsgn += mulscale22(r[zz].y,dy);
			if (dz != 0) zzsgn += mulscale22(r[zz].z,dz);
			if (zsgn < 0)
			{
				copybuf(&r[z],&r[numpoints],3);
				splittag[numpoints] = splittag[z];
				point2[numpoints] = numpoints+1;
				numpoints++;
			}
			else if (zsgn > 0)
			{
				copybuf(&r[z],&r[numpoints2],3);
				splittag[numpoints2] = splittag[z];
				point2[numpoints2] = numpoints2-1;
				numpoints2--;
			}
			if (((zsgn^zzsgn) < 0) || (((-zsgn)^(-zzsgn)) < 0))
			{
				i = divscale28(zsgn,zsgn-zzsgn);
				intx = r[z].x + mulscale28(r[zz].x-r[z].x,i);
				inty = r[z].y + mulscale28(r[zz].y-r[z].y,i);
				intz = r[z].z + mulscale28(r[zz].z-r[z].z,i);
				if ((zsgn^zzsgn) < 0)
				{
					r[numpoints].x = intx;
					r[numpoints].y = inty;
					r[numpoints].z = intz;
					point2[numpoints] = numpoints+1;
					if (zsgn < 0)
						{ split[splitcnt++] = numpoints; splittag[numpoints] = p; }
					else
						splittag[numpoints] = splittag[z];
					numpoints++;
				}
				if (((-zsgn)^(-zzsgn)) < 0)
				{
					r[numpoints2].x = intx;
					r[numpoints2].y = inty;
					r[numpoints2].z = intz;
					point2[numpoints2] = numpoints2-1;
					if (zsgn > 0)
						{ split[splitcnt2--] = numpoints2; splittag[numpoints2] = p; }
					else
						splittag[numpoints2] = splittag[z];
					numpoints2--;
				}
			}
			z = zz;
		} while (point2[z] >= 0);

		if (numpoints >= start2+3)
			{ point2[numpoints-1] = start2; start2 = numpoints; }
		else
			numpoints = start2;

		if (numpoints2 <= start22-3)
			{ point2[numpoints2+1] = start22; start22 = numpoints2; }
		else
			numpoints2 = start22;

		z = pend;
		while ((z > pstart) && (point2[z] < 0)) z--;
	} while (z > pstart);

	for(z=pstart;z<=pend;z++) point2[z] &= 0x7fff;

	/*
	dx = klabs(dx); dy = klabs(dy); dz = klabs(dz);
	if ((dx >= dy) && (dx >= dz))
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(r[s1].x-r[s4].x)+klabs(r[s3].x-r[s2].x) < klabs(r[s1].x-r[s2].x) + klabs(r[s3].x-r[s4].x))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
		for(z=1;z<MAXSPLITCNT-splitcnt2-1;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[MAXSPLITCNT-1-z]; s2 = point2[s1]; s3 = split[MAXSPLITCNT-1-zz]; s4 = point2[s2];
				if (klabs(r[s1].x-r[s4].x)+klabs(r[s3].x-r[s2].x) < klabs(r[s1].x-r[s2].x) + klabs(r[s3].x-r[s4].x))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}
	else if (dy >= dz)
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(r[s1].y-r[s4].y)+klabs(r[s3].y-r[s2].y) < klabs(r[s1].y-r[s2].y) + klabs(r[s3].y-r[s4].y))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
		for(z=1;z<MAXSPLITCNT-splitcnt2-1;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[MAXSPLITCNT-1-z]; s2 = point2[s1]; s3 = split[MAXSPLITCNT-1-zz]; s4 = point2[s2];
				if (klabs(r[s1].y-r[s4].y)+klabs(r[s3].y-r[s2].y) < klabs(r[s1].y-r[s2].y) + klabs(r[s3].y-r[s4].y))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}
	else
	{
		for(z=1;z<splitcnt;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[z]; s2 = point2[s1]; s3 = split[zz]; s4 = point2[s3];
				if (klabs(r[s1].z-r[s4].z)+klabs(r[s3].z-r[s2].z) < klabs(r[s1].z-r[s2].z) + klabs(r[s3].z-r[s4].z))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
		for(z=1;z<MAXSPLITCNT-splitcnt2-1;z++)
			for(zz=0;zz<z;zz++)
			{
				s1 = split[MAXSPLITCNT-1-z]; s2 = point2[s1]; s3 = split[MAXSPLITCNT-1-zz]; s4 = point2[s2];
				if (klabs(r[s1].z-r[s4].z)+klabs(r[s3].z-r[s2].z) < klabs(r[s1].z-r[s2].z) + klabs(r[s3].z-r[s4].z))
					{ i = point2[s1]; point2[s1] = point2[s3]; point2[s3] = i; }
			}
	}*/

	numpolys++; poly[numpolys] = numpoints;

	zz = (MAXPOINTS-1) + numpoints;
	for(z=MAXPOINTS-1;z>numpoints2;z--)
	{
		copybuf(&r[z],&r[numpoints],3);
		splittag[numpoints] = splittag[z];
		point2[numpoints] = zz-point2[z];
		numpoints++;
	}
	numpolys++; poly[numpolys] = numpoints;
}

krand()
{
	randomseed = ((randomseed*21+1)&65535);
	return(randomseed);
}

drawline256 (long x1, long y1, long x2, long y2, char col)
{
	long dx, dy, i, j, p, inc, plc, daend;
	long wx1 = 0L, wy1 = 0L, wx2 = (xdim<<12), wy2 = (ydim<<12);

	col = palookup[0][col];

	x1 >>= 4; y1 >>= 4; x2 >>= 4; y2 >>= 4;

	dx = x2-x1; dy = y2-y1;
	if (dx >= 0)
	{
		if ((x1 >= wx2) || (x2 < wx1)) return;
		if (x1 < wx1) y1 += scale(wx1-x1,dy,dx), x1 = wx1;
		if (x2 > wx2) y2 += scale(wx2-x2,dy,dx), x2 = wx2;
	}
	else
	{
		if ((x2 >= wx2) || (x1 < wx1)) return;
		if (x2 < wx1) y2 += scale(wx1-x2,dy,dx), x2 = wx1;
		if (x1 > wx2) y1 += scale(wx2-x1,dy,dx), x1 = wx2;
	}
	if (dy >= 0)
	{
		if ((y1 >= wy2) || (y2 < wy1)) return;
		if (y1 < wy1) x1 += scale(wy1-y1,dx,dy), y1 = wy1;
		if (y2 > wy2) x2 += scale(wy2-y2,dx,dy), y2 = wy2;
	}
	else
	{
		if ((y2 >= wy2) || (y1 < wy1)) return;
		if (y2 < wy1) x2 += scale(wy1-y2,dx,dy), y2 = wy1;
		if (y1 > wy2) x1 += scale(wy2-y1,dx,dy), y1 = wy2;
	}

	if (klabs(dx) >= klabs(dy))
	{
		if (dx == 0) return;
		if (dx < 0)
		{
			i = x1; x1 = x2; x2 = i;
			i = y1; y1 = y2; y2 = i;
		}

		inc = divscale12(dy,dx);
		plc = y1+mulscale12((2047-x1)&4095,inc);
		i = ((x1+2048)>>12); daend = ((x2+2048)>>12);
		for(;i<daend;i++)
		{
			drawpixel(ylookup[plc>>12]+i+frameplace,col);
			plc += inc;
		}
	}
	else
	{
		if (dy < 0)
		{
			i = x1; x1 = x2; x2 = i;
			i = y1; y1 = y2; y2 = i;
		}

		inc = divscale12(dx,dy);
		plc = x1+mulscale12((2047-y1)&4095,inc);
		i = ((y1+2048)>>12); daend = ((y2+2048)>>12);
		p = ylookup[i]+frameplace;
		for(;i<daend;i++)
		{
			drawpixel((plc>>12)+p,col);
			plc += inc; p += xdim;
		}
	}
}

printext256(long xpos, long ypos, short col, short backcol, char name[82], char fontsize)
{
	long stx, zx, stupidbitoffset, x, y, charxsiz;
	short dat, dacol;
	char *fontptr;

	stx = xpos;

	if (fontsize == 1)
	{
		fontptr = smalltextfont;
		charxsiz = 4;
		stupidbitoffset = 1;
	}
	else
	{
		fontptr = textfont;
		charxsiz = 8;
		stupidbitoffset = 0;
	}

	for(zx=0;zx<strlen(name);zx++)
	{
		dat = name[zx];

		for(y=0;y<8;y++)
			for(x=0;x<charxsiz;x++)
			{
				if ((fontptr[(dat<<3)+y]&(1<<(7-stupidbitoffset-x))) > 0)
					dacol = (short)col;
				else if (backcol >= 0)
					dacol = (short)backcol;
				else
					dacol = -1;

				if (dacol >= 0)
					drawpixel(ylookup[y+ypos]+(x+stx-stupidbitoffset)+frameplace,(long)dacol);
			}

		stx += charxsiz;
	}
}

loadtile (short tnum)
{
	char *ptr;
	long i, j, k, l, xsiz, ysiz;

	if ((tilesizx[tnum] <= 0) || (tilesizy[tnum] <= 0)) return;
	if (tilesizy[tnum] > 256) return;

	i = tilefilenum[tnum];
	if (i != artfilnum)
	{
		if (artfil != -1) close(artfil);
		artfilnum = i;
		artfilplc = 0L;

		artfilename[7] = (i%10)+48;
		artfilename[6] = ((i/10)%10)+48;
		artfilename[5] = ((i/100)%10)+48;
		artfil = open(artfilename,O_BINARY|O_RDONLY,S_IREAD);
	}

	xsiz = tilesizx[tnum];
	ysiz = tilesizy[tnum]; if (ysiz > 256) ysiz = 256;

	if (waloff[tnum] == 0) cachebox(&waloff[tnum],xsiz,ysiz);

	if (artfilplc != tilefileoffs[tnum])
	{
		artfilplc = tilefileoffs[tnum];
		lseek(artfil,artfilplc,SEEK_SET);
	}
	ptr = (char *)waloff[tnum];

	l = 16;
	if (ysiz <= 128)
	{
		l = 32;
		if (ysiz <= 64)
		{
			l = 64;
			if (ysiz <= 32) l = 128;
		}
	}
	for(i=0;i<xsiz;i+=l)
	{
		j = min(xsiz-i,l);
		read(artfil,tempbuf,j*ysiz);

		k = FP_OFF(tempbuf);
		while (j > 0)
		{
			copybufbyte(k,ptr,ysiz);
			j--; k += ysiz; ptr += 256;
		}
	}
	artfilplc += xsiz*ysiz;
}

loadpics(char *filename)
{
	long i, j, k, fil, offscount, siz, templongbuf[4];

	strcpy(artfilename,filename);

	for(i=0;i<MAXTILES;i++)
		{ waloff[i] = 0; tilesizx[i] = 0L; tilesizy[i] = 0L; picanm[i] = 0L; }

	artsize = 0L;

	numtilefiles = 0;
	do
	{
		k = numtilefiles;

		artfilename[7] = (k%10)+48;
		artfilename[6] = ((k/10)%10)+48;
		artfilename[5] = ((k/100)%10)+48;
		if ((fil = open(artfilename,O_BINARY|O_RDONLY,S_IREAD)) != -1)
		{
			read(fil,templongbuf,16);
			artversion = templongbuf[0]; if (artversion != 1) return(-1);
			numtiles = templongbuf[1];
			localtilestart[k] = templongbuf[2];
			localtileend[k] = templongbuf[3];
			read(fil,tempshort,(localtileend[k]-localtilestart[k]+1)<<1);
			for(i=localtileend[k]-localtilestart[k];i>=0;i--)
				tilesizx[i+localtilestart[k]] = tempshort[i];
			read(fil,tempshort,(localtileend[k]-localtilestart[k]+1)<<1);
			for(i=localtileend[k]-localtilestart[k];i>=0;i--)
				tilesizy[i+localtilestart[k]] = tempshort[i];
			read(fil,&picanm[localtilestart[k]],(localtileend[k]-localtilestart[k]+1)<<2);

			offscount = 4+4+4+4+((localtileend[k]-localtilestart[k]+1)<<3);
			for(i=localtilestart[k];i<=localtileend[k];i++)
			{
				tilefilenum[i] = k;
				tilefileoffs[i] = offscount;
				j = tilesizx[i]*tilesizy[i];
				offscount += j; artsize += j;
			}
			close(fil);

			numtilefiles++;
		}
	}
	while (k != numtilefiles);

		//150% of art size
	i = ((artsize+(artsize>>1))&0xffffff00);
	while ((pic = (char *)malloc(i)) == NULL) i -= 65536;
	initcache(FP_OFF(pic),i);

	for(i=0;i<MAXTILES;i++)
	{
		j = 15;
		while ((j > 0) && ((1<<j) > tilesizx[i])) j--;
		if ((1<<j) != tilesizx[i]) j++;    //THIS FOR NONREPEATING TEXTURES!
		picsiz[i] = ((char)j);
		j = 15;
		while ((j > 0) && ((1<<j) > tilesizy[i])) j--;
		if ((1<<j) != tilesizy[i]) j++;    //THIS FOR NONREPEATING TEXTURES!
		picsiz[i] += ((char)(j<<4));
	}

	picsloaded = 1;

	artfil = -1;
	artfilnum = -1;
	artfilplc = 0L;

	return(0);
}

unloadpics()
{
	if (artfil != -1) close(artfil);
	free(pic);
	picsloaded = 0;
}

animateoffs(long tilenum)
{
	long i, k, offs;

	offs = 0;
	i = (baktotalclock>>((picanm[tilenum]>>24)&15));
	if ((picanm[tilenum]&63) > 0)
	{
		switch(picanm[tilenum]&192)
		{
			case 64:
				k = (i%((picanm[tilenum]&63)<<1));
				if (k < (picanm[tilenum]&63))
					offs = k;
				else
					offs = (((picanm[tilenum]&63)<<1)-k);
				break;
			case 128:
				offs = (i%((picanm[tilenum]&63)+1));
					break;
			case 192:
				offs = -(i%((picanm[tilenum]&63)+1));
		}
	}
	return(offs);
}

setvesamode(short vesamode)
{
	static struct rminfo
	{
		long EDI, ESI, EBP, ESP, EBX, EDX, ECX, EAX;
		short flags, ES, DS, FS, GS, IP, CS, SP, SS;
	} RMI;

	union REGS regs;
	struct SREGS sregs;
	short selector, segment;
	short *ptr, gran;

	memset(&sregs,0,sizeof(sregs));
	regs.w.ax = 0x0100;
	regs.w.bx = (256>>4);             //Number of paragraphs
	int386x(0x31,&regs,&regs,&sregs); // DPMI call 100h allocates DOS memory
	segment = regs.w.ax;
	selector = regs.w.dx;

	memset(&RMI,0,sizeof(RMI));    // Set up real-mode call structure
	RMI.EAX = 0x00004f01;
	RMI.ECX = (long)vesamode;
	RMI.ES = segment;
	RMI.EDI = 0;

	regs.w.ax = 0x0300;            // Use DMPI call 300h to issue the DOS interrupt
	regs.h.bl = 0x10;
	regs.h.bh = 0;
	regs.w.cx = 0;
	sregs.es = FP_SEG(&RMI);
	regs.x.edi = FP_OFF(&RMI);
	int386x(0x31,&regs,&regs,&sregs);

	if ((RMI.EAX&65535) != 0x004f)
		return(-1);

	ptr = (short *)((((long)segment)<<4)+4);
	gran = 64 / (*ptr);
	vesapageshift = 0;
	while (gran > 1)
	{
		vesapageshift++;
		gran >>= 1;
	}

	if (vesasetmode(vesamode) != 0x4f)
		return(-1);

	return(0);
}

gettile(short dapicnum)
{
	char *ptr;
	long xpicnum, ypicnum, xypicnum;
	short topleftpic;

	xpicnum = (xdim>>6);
	ypicnum = (ydim>>6);
	xypicnum = xpicnum*ypicnum;

	topleftpic = (dapicnum/xypicnum)*xypicnum;
	drawtilescreen(topleftpic);

	xormousecursor();
	while (1)
	{
		obstatus = bstatus, getmousevalues(&mousx,&mousy,&bstatus);
		if ((mousx|mousy) != 0)
		{
			xormousecursor();
			searchx += mousx;
			searchy += mousy;
			if (searchx < 4) searchx = 4;
			if (searchx >= xdim-4) searchx = xdim-4-1;
			if (searchy < 4) searchy = 4;
			if (searchy >= ydim-4) searchy = ydim-4-1;
			xormousecursor();
		}

		if (keystatus[0xcb] > 0)
		{
			keystatus[0xcb] = 0;
			xormousecursor();
			searchx -= 64;
			if (searchx < 4) searchx = 4;
			xormousecursor();
		}
		if (keystatus[0xcd] > 0)
		{
			keystatus[0xcd] = 0;
			xormousecursor();
			searchx += 64;
			if (searchx >= xdim-4) searchx = xdim-4-1;
			xormousecursor();
		}
		if (keystatus[0xc8] > 0)
		{
			keystatus[0xc8] = 0;
			xormousecursor();
			searchy -= 64;
			if (searchy < 4) searchy = 4;
			xormousecursor();
		}
		if (keystatus[0xd0] > 0)
		{
			keystatus[0xd0] = 0;
			xormousecursor();
			searchy += 64;
			if (searchx >= xdim-4) searchx = xdim-4-1;
			xormousecursor();
		}
		if (keystatus[0xc9] > 0)  //PGUP
		{
			if (topleftpic > 0)
			{
				topleftpic -= xypicnum;
				drawtilescreen(topleftpic);
				xormousecursor();
			}
			keystatus[0xc9] = 0;
		}
		if (keystatus[0xd1] > 0)  //PGDN
		{
			if (topleftpic < MAXTILES-xypicnum)
			{
				topleftpic += xypicnum;
				drawtilescreen(topleftpic);
				xormousecursor();
			}
			keystatus[0xd1] = 0;
		}

		if (keystatus[1] > 0)
		{
			keystatus[1] = 0;
			return(dapicnum);
		}
		if (keystatus[0x1c] > 0)
		{
			keystatus[0x1c] = 0;
			if ((searchx < (xpicnum<<6)) && (searchy < (ypicnum<<6)))
				dapicnum = topleftpic+(searchy>>6)*xpicnum+(searchx>>6);
			return(dapicnum);
		}

	}
}

drawtilescreen(short topleftpic)
{
	char *ptr;
	long i, j, xsiz, ysiz, ptrinc, x, y, p;
	short dapic;


	clearbuf(frameplace,(xdim*ydim)>>2,0x06070807);

		//Draw tiles on screen
	dapic = topleftpic;
	for(j=0;j<=ydim-64;j+=64)
		for(i=0;i<=xdim-64;i+=64)
		{
			xsiz = (long)tilesizx[dapic];
			ysiz = (long)tilesizy[dapic];
			if (waloff[dapic] == 0) loadtile(dapic);

			if ((xsiz > 0) && (ysiz > 0) && (waloff[dapic] != 0))
			{
				ptrinc = 1;
				while ((xsiz > 64) || (ysiz > 64))
				{
					xsiz >>= 1;
					ysiz >>= 1;
					ptrinc <<= 1;
				}
				if (xsiz > 64) xsiz = 64;
				if (ysiz > 64) ysiz = 64;

				for(x=0;x<xsiz;x++)
				{
					ptr = (char *)(waloff[dapic]+((x*ptrinc)<<8));
					p = frameplace+ylookup[j]+(x+i);
					for(y=0;y<ysiz;y++)
					{
						drawpixel(p,*ptr);
						p += xdim;
						ptr += ptrinc;
					}
				}
			}
			dapic++;
		}

	nextpage();

	if (option[0] >= 2)
	{
		visualpage = ((visualpage+3)&3);
		if (option[0] == 2) koutp(0x3cd,visualpage+(visualpage<<4));
		if (option[0] == 3) koutp(0x3ce,0x9), koutp(0x3cf,visualpage<<4);
		if (option[0] == 4) koutp(0x3d4,0x35), koutp(0x3d5,visualpage);
	}
}

xormousecursor()
{
	long i, p;

	for(i=1;i<=4;i++)
	{
		p = ylookup[searchy-i] + (searchx+0);
		if (option[0] <= 1) vesasetpage(p>>16); p = 0xa0000+(p&65535); drawpixel(p,readpixel(p)^144);
		p = ylookup[searchy+0] + (searchx-i);
		if (option[0] <= 1) vesasetpage(p>>16); p = 0xa0000+(p&65535); drawpixel(p,readpixel(p)^144);
		p = ylookup[searchy+0] + (searchx+i);
		if (option[0] <= 1) vesasetpage(p>>16); p = 0xa0000+(p&65535); drawpixel(p,readpixel(p)^144);
		p = ylookup[searchy+i] + (searchx+0);
		if (option[0] <= 1) vesasetpage(p>>16); p = 0xa0000+(p&65535); drawpixel(p,readpixel(p)^144);
	}
}

initmouse()
{
	if ((setupandgetmouseirq()&65535) == 0) return(0);
	switch(mouseirq)
	{
		case 3: mouseport = 0x2f8; break;
		case 4: mouseport = 0x3f8; break;
		default: mouseirq = 4; mouseport = 0x3f8; break;
	}
	oldmousehandler = _dos_getvect(mouseirq+0x8);
	_disable(); _dos_setvect(mouseirq+0x8, mousehandler); _enable();
	return(1);
}

uninitmouse()
{
	if (mouseport <= 0) return;
	_disable(); _dos_setvect(mouseirq+0x8, oldmousehandler); _enable();
}

void __interrupt __far mousehandler()
{
	mousetemp = kinp(mouseport); if (mousetemp&0x40) mousebytecnt = 0;
	mousebyte[mousebytecnt++] = mousetemp;
	if (mousebytecnt > 2)
	{
		mousebytecnt = 0;
		intmousx += ((signed char)((mousebyte[0]<<6)&192)+(mousebyte[1]&63));
		intmousy += ((signed char)((mousebyte[0]<<4)&192)+(mousebyte[2]&63));
		intbstatus = mouselookup[(mousebyte[0]>>4)&3];
	}
	koutp(0x20,0x20);
}

getmousevalues(short *mousx, short *mousy, short *bstatus)
{
	if (mouseport <= 0) { *mousx = 0; *mousy = 0; *bstatus = 0; return; }
	*mousx = intmousx; intmousx -= *mousx;
	*mousy = intmousy; intmousy -= *mousy;
	*bstatus = intbstatus;
}

searchit()
{
	long xv, yv, zv, xp, yp, zp, tx, ty, tz, xd, yd, zd;

	searchitx = posx;
	searchity = posy;
	searchitz = posz;

	xv = ((searchx-(xdim>>1))<<8);
	yv = ((searchy-(ydim>>1))<<8);
	zv = (xdim<<7);
	backrotate3d(xv,yv,zv);

	searchcube = (posx>>10)*BSIZ*BSIZ+(posy>>10)*BSIZ+(posz>>10);
	if (xv < 0) { xp = (posx&1023); xd = -BSIZ*BSIZ; xv = -xv; }
			 else { xp = (-posx&1023); xd = BSIZ*BSIZ; }
	if (yv < 0) { yp = (posy&1023); yd = -BSIZ; yv = -yv; }
			 else { yp = (-posy&1023); yd = BSIZ; }
	if (zv < 0) { zp = (posz&1023); zd = -1; zv = -zv; }
			 else { zp = (-posz&1023); zd = 1; }
	if (xv <= 2) xv = 0x001fffff; else xv = divscale32(1L,xv);
	if (yv <= 2) yv = 0x001fffff; else yv = divscale32(1L,yv);
	if (zv <= 2) zv = 0x001fffff; else zv = divscale32(1L,zv);
	tx = mulscale10(xp,xv); ty = mulscale10(yp,yv); tz = mulscale10(zp,zv);
	while (1)
	{
		if (tx < ty)
		{
			if (tx < tz)
			{
				searchcube += xd; if (board[searchcube] >= 0) { searchcubak = searchcube-xd; searchedge = 0+((xd<0)*3); return; }
				ty -= tx; tz -= tx; tx = xv;
			}
			else
			{
				searchcube += zd; if (board[searchcube] >= 0) { searchcubak = searchcube-zd; searchedge = 2+((zd<0)*3); return; }
				tx -= tz; ty -= tz; tz = zv;
			}
		}
		else
		{
			if (ty < tz)
			{
				searchcube += yd; if (board[searchcube] >= 0) { searchcubak = searchcube-yd; searchedge = 1+((yd<0)*3); return; }
				tx -= ty; tz -= ty; ty = yv;
			}
			else
			{
				searchcube += zd; if (board[searchcube] >= 0) { searchcubak = searchcube-zd; searchedge = 2+((zd<0)*3); return; }
				tx -= tz; ty -= tz; tz = zv;
			}
		}
	}
}

clipmove (long *x, long *y, long *z, long xv, long yv, long zv)
{
	if (rayscan(x,y,z,&xv,&yv,&zv) != 255)
		rayscan(x,y,z,&xv,&yv,&zv);

	*x = min(max(*x,1024+128),(BSIZ<<10)-(1024+128));
	*y = min(max(*y,1024+128),(BSIZ<<10)-(1024+128));
	*z = min(max(*z,1024+128),(BSIZ<<10)-(1024+128));
}

rayscan(long *x, long *y, long *z, long *xv, long *yv, long *zv)
{
	long i, j, t, gridx, gridy, gridz, xdir, ydir, zdir;
	long nintx, ninty, nintz, intx, inty, intz;
	char bad, lasthit;

	lasthit = 255;
	intx = (*x) + (*xv);
	inty = (*y) + (*yv);
	intz = (*z) + (*zv);

	if (*xv != 0)
	{
		xdir = ksgn(*xv);
		if (xdir < 0)
			{ gridx = (((*x)-WALLDIST)>>10)-1; nintx = (gridx<<10)+1024+WALLDIST; }
		else
			{ gridx = (((*x)+WALLDIST-1)>>10)+1; nintx = (gridx<<10)-WALLDIST; }
		j = xdir*BSIZ*BSIZ;

		while (klabs(nintx-(*x)) <= klabs(intx-(*x)))
		{
			i = divscale16(intx-(*x),*xv);
			ninty = *y + mulscale16(*yv,i);
			nintz = *z + mulscale16(*zv,i);

			bad = 0;
			i = gridx*BSIZ*BSIZ+((ninty-WALLDIST)>>10)*BSIZ+((nintz-WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = gridx*BSIZ*BSIZ+((ninty+WALLDIST)>>10)*BSIZ+((nintz-WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = gridx*BSIZ*BSIZ+((ninty-WALLDIST)>>10)*BSIZ+((nintz+WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = gridx*BSIZ*BSIZ+((ninty+WALLDIST)>>10)*BSIZ+((nintz+WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			if (bad != 0)
			{
				lasthit = 0;
				if (klabs(nintx-(*x)) < klabs(intx-(*x))) intx = nintx;
				if (klabs(ninty-(*y)) < klabs(inty-(*y))) inty = ninty;
				if (klabs(nintz-(*z)) < klabs(intz-(*z))) intz = nintz;
				break;
			}
			gridx += xdir; nintx += (xdir<<10);
		}
	}
	if (*yv != 0)
	{
		ydir = ksgn(*yv);
		if (ydir < 0)
			{ gridy = (((*y)-WALLDIST)>>10)-1; ninty = (gridy<<10)+1024+WALLDIST; }
		else
			{ gridy = (((*y)+WALLDIST-1)>>10)+1; ninty = (gridy<<10)-WALLDIST; }
		j = ydir*BSIZ;

		while (klabs(ninty-(*y)) <= klabs(inty-(*y)))
		{
			i = divscale16(inty-(*y),*yv);
			nintx = *x + mulscale16(*xv,i);
			nintz = *z + mulscale16(*zv,i);

			bad = 0;
			i = ((nintx-WALLDIST)>>10)*BSIZ*BSIZ+gridy*BSIZ+((nintz-WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = ((nintx+WALLDIST)>>10)*BSIZ*BSIZ+gridy*BSIZ+((nintz-WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = ((nintx-WALLDIST)>>10)*BSIZ*BSIZ+gridy*BSIZ+((nintz+WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = ((nintx+WALLDIST)>>10)*BSIZ*BSIZ+gridy*BSIZ+((nintz+WALLDIST)>>10); if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			if (bad != 0)
			{
				lasthit = 1;
				if (klabs(nintx-(*x)) < klabs(intx-(*x))) intx = nintx;
				if (klabs(ninty-(*y)) < klabs(inty-(*y))) inty = ninty;
				if (klabs(nintz-(*z)) < klabs(intz-(*z))) intz = nintz;
				break;
			}
			gridy += ydir; ninty += (ydir<<10);
		}
	}
	if (*zv != 0)
	{
		zdir = ksgn(*zv);
		if (zdir < 0)
			{ gridz = (((*z)-WALLDIST)>>10)-1; nintz = (gridz<<10)+1024+WALLDIST; }
		else
			{ gridz = (((*z)+WALLDIST-1)>>10)+1; nintz = (gridz<<10)-WALLDIST; }
		j = zdir;

		while (klabs(nintz-(*z)) <= klabs(intz-(*z)))
		{
			i = divscale16(intz-(*z),*zv);
			nintx = *x + mulscale16(*xv,i);
			ninty = *y + mulscale16(*yv,i);

			bad = 0;
			i = ((nintx-WALLDIST)>>10)*BSIZ*BSIZ+((ninty-WALLDIST)>>10)*BSIZ+gridz; if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = ((nintx+WALLDIST)>>10)*BSIZ*BSIZ+((ninty-WALLDIST)>>10)*BSIZ+gridz; if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = ((nintx-WALLDIST)>>10)*BSIZ*BSIZ+((ninty+WALLDIST)>>10)*BSIZ+gridz; if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			i = ((nintx+WALLDIST)>>10)*BSIZ*BSIZ+((ninty+WALLDIST)>>10)*BSIZ+gridz; if ((board[i] >= 0) && (board[i-j] < 0)) bad = 1;
			if (bad != 0)
			{
				lasthit = 2;
				if (klabs(nintx-(*x)) < klabs(intx-(*x))) intx = nintx;
				if (klabs(ninty-(*y)) < klabs(inty-(*y))) inty = ninty;
				if (klabs(nintz-(*z)) < klabs(intz-(*z))) intz = nintz;
				break;
			}
			gridz += zdir; nintz += (zdir<<10);
		}
	}

	if (lasthit == 0) *xv = 0; else *xv = (*x) + (*xv) - intx;
	if (lasthit == 1) *yv = 0; else *yv = (*y) + (*yv) - inty;
	if (lasthit == 2) *zv = 0; else *zv = (*z) + (*zv) - intz;
	*x = intx; *y = inty; *z = intz;
	return(lasthit);
}
