// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#define MAXSECTORS 1024
#define MAXWALLS 4096
#define MAXSPRITES 4096
#define MAXSTATUS 1024
#define MAXTILES 4096

static struct sectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingosc;
	signed char ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorosc;
	signed char floorshade;
	long floorz;
	short floorpicnum, floorheinum;
	long tag;
};

static struct walltype
{
	long x, y;
	short point2;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	long tag;
};

static struct spritetype
{
	long x, y, z;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
	short sectnum, statnum;
	long tag;
};

static struct newsectortype
{
	short unsigned int wallptr, wallnum;
	char ceilingstat, ceilingpanning;
	signed char ceilingshade;
	long ceilingz;
	short ceilingpicnum, ceilingheinum;
	char floorstat, floorpanning;
	signed char floorshade;
	long floorz;
	short floorpicnum, floorheinum;
	long tag;
};

static struct newwalltype
{
	long x, y;
	short point2;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat, panning;
	short picnum, overpicnum;
	short nextsector1, nextwall1;
	short nextsector2, nextwall2;
	long tag;
};

static struct newspritetype
{
	long x, y, z;
	char cstat;
	signed char shade;
	unsigned char xrepeat, yrepeat;
	short picnum, ang, vel, zvel, owner;
	short sectnum, statnum;
	long tag;
	char *extra;
};

//**************************************************************************

static struct sectortype sector[MAXSECTORS];
static struct walltype wall[MAXWALLS];
static struct spritetype sprite[MAXSPRITES];

static struct newsectortype newsector[MAXSECTORS];
static struct newwalltype newwall[MAXWALLS];
static struct newspritetype newsprite[MAXSPRITES];

static long posx, posy, posz, mapversion;
static short int ang, cursectnum, numsectors, numwalls, numsprites;

static long artversion;
static long numtiles;
static long localtilestart;
static long localtileend;
static short tilesizx[MAXTILES];
static short tilesizy[MAXTILES];

main(int argc, char **argv)
{
	char ch, filename[80];
	long i, j;

	if (argc != 2)
	{
		printf("CONVMAP3 [filename]\n");
		printf("This CONVMAP3.EXE converts map version 2 to map version 3\n");
		exit(0);
	}
	strcpy(filename,argv[1]);
	if(strchr(filename,'.') == 0)
		strcat(filename,".map");

	printf("Converting from map version 2 to map version 3...\n");
	printf("Adding new fields to structures...\n");

	if ((i = loadoldboard(filename)) < 0)
	{
		if (i == -1) printf("%s not found.\n",filename);
		if (i == -2) printf("Cannot convert %s.\n",filename);
		exit(0);
	}

	for(i=0;i<numsectors;i++)
	{
		newsector[i].wallptr = sector[i].wallptr;
		newsector[i].wallnum = sector[i].wallnum;
		newsector[i].ceilingstat = sector[i].ceilingstat;
		newsector[i].ceilingpanning = 0;
		newsector[i].ceilingshade = sector[i].ceilingshade;
		newsector[i].ceilingz = sector[i].ceilingz;
		newsector[i].ceilingpicnum = sector[i].ceilingpicnum;
		newsector[i].ceilingheinum = sector[i].ceilingheinum;
		newsector[i].floorstat = sector[i].floorstat;
		newsector[i].floorpanning = 0;
		newsector[i].floorshade = sector[i].floorshade;
		newsector[i].floorz = sector[i].floorz;
		newsector[i].floorpicnum = sector[i].floorpicnum;
		newsector[i].floorheinum = sector[i].floorheinum;
		newsector[i].tag = sector[i].tag;
	}
	for(i=0;i<numwalls;i++)
	{
		newwall[i].x = wall[i].x;
		newwall[i].y = wall[i].y;
		newwall[i].point2 = wall[i].point2;
		newwall[i].cstat = wall[i].cstat;
		newwall[i].shade = wall[i].shade;
		newwall[i].xrepeat = wall[i].xrepeat;
		newwall[i].yrepeat = wall[i].yrepeat;
		newwall[i].panning = 0;
		newwall[i].picnum = wall[i].picnum;
		newwall[i].overpicnum = -1;
		newwall[i].nextsector1 = wall[i].nextsector1;
		newwall[i].nextwall1 = wall[i].nextwall1;
		newwall[i].nextsector2 = wall[i].nextsector2;
		newwall[i].nextwall2 = wall[i].nextwall2;
		newwall[i].tag = wall[i].tag;
	}
	for(i=0;i<numsprites;i++)
	{
		newsprite[i].x = sprite[i].x;
		newsprite[i].y = sprite[i].y;
		newsprite[i].z = sprite[i].z;
		newsprite[i].cstat = sprite[i].cstat;
		newsprite[i].shade = sprite[i].shade;
		newsprite[i].xrepeat = sprite[i].xrepeat;
		newsprite[i].yrepeat = sprite[i].yrepeat;
		newsprite[i].picnum = sprite[i].picnum;
		newsprite[i].ang = sprite[i].ang;
		newsprite[i].vel = sprite[i].vel;
		newsprite[i].zvel = sprite[i].zvel;
		newsprite[i].owner = sprite[i].owner;
		newsprite[i].sectnum = sprite[i].sectnum;
		newsprite[i].statnum = sprite[i].statnum;
		newsprite[i].tag = sprite[i].tag;
		newsprite[i].extra = 0;
	}

	if (fakeloadpics("tiles000.art") == 0)
	{
		for(i=0;i<numwalls;i++)
			if (tilesizy[newwall[i].picnum] > 0)
			{
				j = (long)newwall[i].xrepeat;
				j *= 64;
				j /= tilesizy[newwall[i].picnum];
				newwall[i].xrepeat = (unsigned char)j;
			}
	}

	for(i=0;i<numwalls;i++)
		if (newwall[i].nextsector1 >= 0)
		{
			if ((newwall[i].cstat&4) > 0)
			{
				newwall[i].cstat &= (255-4);
			}
		}

	printf("The coordinate system was screwed up in the old version of 2D EDIT MODE.\n");
	printf("I had the x's and y's reversed and the y's were negative.\n");
	printf("Anyway, press Y here only if you want the board orientation to look the same\n");
	printf("as the old version of 2D EDIT MODE.  Warning: You should press N if you have\n");
	printf("floors that MUST face a certain direction such as a football field or a roadway.\n");
	printf("(Y/N)\n");
	do
	{
		ch = getch();
	} while ((ch != 'Y') && (ch != 'y') && (ch != 'N') && (ch != 'n'));
	if ((ch == 'y') || (ch == 'Y'))
	{
		printf("Fixing coordinates so they look the same in 2D EDIT MODE.\n");
		j = posx;              //Board rotation!
		posx = posy;
		posy = 65536-j;
		ang += 1536;

		for(i=0;i<numwalls;i++)
		{
			j = newwall[i].x;
			newwall[i].x = newwall[i].y;
			newwall[i].y = 65536-j;
		}
		for(i=0;i<MAXSPRITES;i++)
			if (newsprite[i].statnum < MAXSTATUS)
			{
				j = newsprite[i].x;
				newsprite[i].x = newsprite[i].y;
				newsprite[i].y = 65536-j;
			}
	}

	mapversion = 3L;
	if (savenewboard(filename) == 0)
		printf("%s successfully converted.\n",filename);
	else
		printf("%s NOT converted for some stupid reason.\n",filename);
}

loadoldboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_RDWR,S_IREAD)) == -1)
		return(-1);

	read(fil,&mapversion,4);
	if (mapversion != 2L) return(-2);
	read(fil,&posx,4);
	read(fil,&posy,4);
	read(fil,&posz,4);
	read(fil,&ang,2);
	read(fil,&cursectnum,2);
	read(fil,&numsectors,2);
	read(fil,&numwalls,2);
	read(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) read(fil,&sector[i],sizeof(struct sectortype));
	for(i=0;i<numwalls;i++) read(fil,&wall[i],sizeof(struct walltype));
	for(i=0;i<numsprites;i++) read(fil,&sprite[i],sizeof(struct spritetype));
	close(fil);

	return(0);
}

savenewboard(char filename[20])
{
	short int fil, i;

	if ((fil = open(filename,O_BINARY|O_TRUNC|O_CREAT|O_WRONLY,S_IWRITE)) == -1)
		return(-1);

	write(fil,&mapversion,4);
	write(fil,&posx,4);
	write(fil,&posy,4);
	write(fil,&posz,4);
	write(fil,&ang,2);
	write(fil,&cursectnum,2);
	write(fil,&numsectors,2);
	write(fil,&numwalls,2);
	write(fil,&numsprites,2);
	for(i=0;i<numsectors;i++) write(fil,&newsector[i],sizeof(struct newsectortype));
	for(i=0;i<numwalls;i++) write(fil,&newwall[i],sizeof(struct newwalltype));
	for(i=0;i<numsprites;i++) write(fil,&newsprite[i],sizeof(struct newspritetype));
	close(fil);

	return(0);
}

fakeloadpics(char filename[20])
{
	long totpicsiz, offscount, siz;
	short int fil, i, j, k, numtilefiles;

	numtilefiles = 0;
	do
	{
		k = numtilefiles;

		filename[7] = (k%10)+48;
		filename[6] = ((k/10)%10)+48;
		filename[5] = ((k/100)%10)+48;
		if ((fil = open(filename,O_BINARY|O_RDONLY,S_IREAD)) != -1)
		{
			read(fil,&artversion,4);
			if (artversion != 1) return(-1);
			read(fil,&numtiles,4);
			read(fil,&localtilestart,4);
			read(fil,&localtileend,4);
			read(fil,&tilesizx[localtilestart],(localtileend-localtilestart+1)<<1);
			read(fil,&tilesizy[localtilestart],(localtileend-localtilestart+1)<<1);

			close(fil);
		}
	}
	while (k != numtilefiles);

	return(0);
}
