// "Build Engine & Tools" Copyright (c) 1993-1997 Ken Silverman
// Ken Silverman's official web site: "http://www.advsys.net/ken"
// See the included license file "BUILDLIC.TXT" for license info.

#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>
#include <dos.h>
#include <malloc.h>

#define MAXPALOOKUPS 64

static long rdist[128], gdist[128], bdist[128];
static long rpal[256], gpal[256], bpal[256];
static long numpalookups;
static char palette[768], paletteloaded;
static char palookup[MAXPALOOKUPS][256];
static char *transluc, translucloaded = 0;
static char palettefilename[13];

#pragma aux setvmode =\
	"int 0x10",\
	parm [eax]\

#pragma aux drawpixel =\
	"mov [edi], al",\
	parm [edi][eax]\

#pragma aux clearbuf =\
	"rep stosd",\
	parm [edi][ecx][eax]\

main(short int argc,char **argv)
{
	char buffer[256], col;
	long fil, i, j;

	if (argc >= 2)
	{
		strcpy(&palettefilename,argv[1]);
		if(strchr(palettefilename,'.') == 0)
			strcat(palettefilename,".dat");
	}
	else
		strcpy(&palettefilename,"palette.dat");

	if (paletteloaded == 0)
	{
		if ((fil = open(palettefilename,O_BINARY|O_RDWR,S_IREAD|S_IWRITE)) == -1)
		{
			printf("%s not found",palettefilename);
			exit(0);
		}

		numpalookups = ((filelength(fil)-768)>>7);
		if ((numpalookups&1) > 0)
		{   //Transluscent palette found in palette.dat
			numpalookups = ((numpalookups-255)>>1);

			if (translucloaded == 0)
				if ((transluc = (char *)malloc(65536L)) != 0)
				{
					//fixtransluscence(FP_OFF(transluc));
					translucloaded = 1;
				}
		}
		else
		{    //No transluscent palette in palette.dat
			numpalookups >>= 1;
		}
		read(fil,&palette[0],768);
		read(fil,&palookup[0][0],(numpalookups<<8));
		if (translucloaded == 1)
		{
			for(i=0;i<255;i++)
			{
				read(fil,&transluc[(i<<8)+i+1],255-i);
				for(j=i+1;j<256;j++)
					transluc[(j<<8)+i] = transluc[(i<<8)+j];
			}
			for(i=0;i<256;i++)
				transluc[(i<<8)+i] = i;
		}
		paletteloaded = 1;
	}

	setvmode(0x13);
	outp(0x3c4,0x4); outp(0x3c5,0x6);
	outp(0x3d4,0x14); outp(0x3d5,0x0);
	outp(0x3d4,0x17); outp(0x3d5,0xe3);

	outpw(0x3d4,9+(0<<8));

	outp(0x3c4,2); outp(0x3c5,15);
	clearbuf(0xa0000,16384,0L);

	outp(0x3c8,0);
	for(i=0;i<768;i++)
		outp(0x3c9,palette[i]);

	for(i=0;i<128;i++)
	{
		rdist[i] = (i-64)*(i-64)*30;
		gdist[i] = (i-64)*(i-64)*59;
		bdist[i] = (i-64)*(i-64)*11;
	}
	for(i=0;i<256;i++)
	{
		rpal[i] = palette[i+i+i+0];
		gpal[i] = palette[i+i+i+1];
		bpal[i] = palette[i+i+i+2];
	}

	for(i=0;i<numpalookups;i++)
		for(j=0;j<256;j++)
		{
			col = getpalookup((char)i,(char)j);
			palookup[i][j] = col;

			outpw(0x3c4,2+(256<<((j+8)&3)));                //Draw top line
			drawpixel(((((i<<1)+0)*320+(j+8))>>2)+0xa0000,(long)col);
			drawpixel(((((i<<1)+1)*320+(j+8))>>2)+0xa0000,(long)col);
		}

	lseek(fil,768L,SEEK_SET);
	write(fil,palookup,numpalookups<<8);

	for(i=0;i<256;i++)
	{
		for(j=0;j<6;j++)
		{
			outpw(0x3c4,2+(256<<((i+8)&3)));                //Draw top line
			drawpixel((((j+132+0)*320+(i+8))>>2)+0xa0000,i);

			outpw(0x3c4,2+(256<<((j+0)&3)));                //Draw left line
			drawpixel((((i+132+8)*320+(j+0))>>2)+0xa0000,i);
		}

		outpw(0x3c4,2+(256<<(i&3)));                       //Draw diagonal line
		drawpixel((((i+132+8)*320+(i+8))>>2)+0xa0000,i);
	}

	for(i=0;i<255;i++)
	{
		for(j=i+1;j<256;j++)
		{
			col = gettrans((char)i,(char)j);
			buffer[j-i-1] = col;

			outpw(0x3c4,2+(256<<((i+8)&3)));
			drawpixel((((j+132+8)*320+(i+8))>>2)+0xa0000,(long)col);
			outpw(0x3c4,2+(256<<((j+8)&3)));                 //Mirror image
			drawpixel((((i+132+8)*320+(j+8))>>2)+0xa0000,(long)col);
		}
		write(fil,&buffer[0],255-i);
	}

	close(fil);

	getch();

	setvmode(0x3);
}

getpalookup(char dashade, char dacol)
{
	long i, ravg, gavg, bavg, triplecol, closest, dist;
	char retcol;

	triplecol = dacol*3;
	dashade = numpalookups-dashade;
	ravg = 64 - ((palette[triplecol+0]*dashade) / numpalookups);
	gavg = 64 - ((palette[triplecol+1]*dashade) / numpalookups);
	bavg = 64 - ((palette[triplecol+2]*dashade) / numpalookups);

	retcol = 0; closest = 0x7fffffff;
	for(i=255;i>=0;i--)
	{
		dist = rdist[rpal[i]+ravg];
		dist += gdist[gpal[i]+gavg];
		dist += bdist[bpal[i]+bavg];
		if (dist < closest) closest = dist, retcol = i;
	}
	return(retcol);
}

gettrans(char dat1, char dat2)
{
	long i, ravg, gavg, bavg, tripledat1, tripledat2, closest, dist;
	char retcol;

	tripledat1 = ((long)dat1<<1)+(long)dat1;
	tripledat2 = ((long)dat2<<1)+(long)dat2;

	ravg = 64 - ((palette[tripledat1+0] + palette[tripledat2+0])>>1);
	gavg = 64 - ((palette[tripledat1+1] + palette[tripledat2+1])>>1);
	bavg = 64 - ((palette[tripledat1+2] + palette[tripledat2+2])>>1);

	retcol = 0; closest = 0x7fffffff;
	for(i=255;i>=0;i--)
	{
		dist = rdist[rpal[i]+ravg];
		dist += gdist[gpal[i]+gavg];
		dist += bdist[bpal[i]+bavg];
		if (dist < closest) closest = dist, retcol = i;
	}
	return(retcol);
}
